from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar

from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
)

from skore._config import get_config

T = TypeVar("T")
DescriptionType = str | Callable[..., str]


def progress_decorator(
    description: DescriptionType,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Decorate class methods to add a progress bar.

    This decorator adds a Rich progress bar to class methods, displaying progress
    during execution. The progress bar automatically disappears after completion.

    Parameters
    ----------
    description : str or callable
        The description of the progress bar. If a callable, it should take the
        self object as an argument and return a string.

    Returns
    -------
    decorator : function
        A decorator that wraps the input function and adds a progress bar to it.
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            self_obj: Any = args[0]

            if hasattr(self_obj, "_parent"):
                # self_obj is an accessor
                self_obj = self_obj._parent

            desc = description(self_obj) if callable(description) else description

            created_progress = False

            if getattr(self_obj, "_progress_info", None) is not None:
                progress = self_obj._progress_info["current_progress"]
            else:
                progress = Progress(
                    SpinnerColumn(),
                    TextColumn("[bold cyan]{task.description}"),
                    BarColumn(
                        complete_style="dark_orange",
                        finished_style="dark_orange",
                        pulse_style="orange1",
                    ),
                    TextColumn("[orange1]{task.percentage:>3.0f}%"),
                    expand=False,
                    transient=True,
                    disable=not get_config()["show_progress"],
                )
                progress.start()
                created_progress = True

            # assigning progress to child reports
            reports_to_cleanup: list[Any] = []
            if hasattr(self_obj, "reports_"):
                for report in self_obj.reports_:
                    if hasattr(report, "_progress_info"):
                        report._progress_info = {"current_progress": progress}
                        reports_to_cleanup.append(report)

            task = progress.add_task(desc, total=None)
            self_obj._progress_info = {
                "current_progress": progress,
                "current_task": task,
            }
            has_errored = False
            try:
                result = func(*args, **kwargs)
                progress.update(
                    task, completed=progress.tasks[task].total, refresh=True
                )
                return result
            except Exception:
                has_errored = True
                raise
            finally:
                if created_progress:
                    if not has_errored:
                        progress.update(
                            task, completed=progress.tasks[task].total, refresh=True
                        )
                    progress.stop()

                # clean up child reports
                for report in reports_to_cleanup:
                    report._progress_info = None

                # clean up to make object pickable
                self_obj._progress_info = None

        return wrapper

    return decorator



"""Types between parts of the sklearn module."""

from collections.abc import Callable, Iterator, Sequence
from dataclasses import dataclass
from typing import Any, Literal, Protocol

from numpy.typing import ArrayLike

PlotBackend = Literal["matplotlib", "plotly"]

MLTask = Literal[
    "binary-classification",
    "classification",
    "clustering",
    "multiclass-classification",
    "multioutput-binary-classification",
    "multioutput-multiclass-classification",
    "multioutput-regression",
    "regression",
    "unknown",
]


class _DefaultType:
    """Sentinel class for default values."""

    def __repr__(self) -> str:
        return "<DEFAULT>"


_DEFAULT = _DefaultType()
PositiveLabel = int | float | bool | str | None | _DefaultType
Aggregate = Literal["mean", "std"] | Sequence[Literal["mean", "std"]]


@dataclass
class YPlotData:
    """Response values, either `y_true` or `y_pred`.

    Used for passing to Display classes.
    """

    estimator_name: str
    split_index: int | None
    y: ArrayLike


ReportType = Literal[
    "cross-validation",
    "estimator",
    "comparison-estimator",
    "comparison-cross-validation",
]


class SKLearnScorer(Protocol):
    """Protocol defining the interface of scikit-learn's _BaseScorer."""

    _score_func: Callable
    _response_method: str | list[str]
    _kwargs: dict[str, Any]


ScoringName = str | None

Scoring = str | Callable | SKLearnScorer


class SKLearnCrossValidator(Protocol):
    """Protocol defining the interface of scikit-learn's cross-validation splitters."""

    def get_n_splits(self, X: Any = None, y: Any = None, groups: Any = None) -> int:
        """Return the number of splitting iterations in the cross-validator."""

    def split(
        self, X: ArrayLike, y: Any = None, groups: Any = None
    ) -> Iterator[tuple[ArrayLike, ArrayLike]]:
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            The target variable for supervised learning problems.

        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """



"""Configure logging and global settings."""

import logging
import warnings

import joblib
from rich.console import Console
from rich.theme import Theme

from skore._config import config_context, get_config, set_config
from skore._externals._sklearn_compat import parse_version
from skore._sklearn import (
    ComparisonReport,
    CrossValidationReport,
    EstimatorReport,
    MetricsSummaryDisplay,
    PrecisionRecallCurveDisplay,
    PredictionErrorDisplay,
    RocCurveDisplay,
    TableReportDisplay,
    find_estimators,
    train_test_split,
)
from skore._sklearn._plot.base import Display
from skore._utils._patch import setup_jupyter_display
from skore._utils._show_versions import show_versions
from skore.project import Project

__all__ = [
    "CrossValidationReport",
    "ComparisonReport",
    "Display",
    "EstimatorReport",
    "PrecisionRecallCurveDisplay",
    "PredictionErrorDisplay",
    "Project",
    "RocCurveDisplay",
    "MetricsSummaryDisplay",
    "show_versions",
    "train_test_split",
    "config_context",
    "get_config",
    "set_config",
    "find_estimators",
    "TableReportDisplay",
]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())  # Default to no output
logger.setLevel(logging.INFO)

# Configure jupyter display for VS Code compatibility
setup_jupyter_display()

skore_console_theme = Theme(
    {
        "repr.str": "cyan",
        "rule.line": "orange1",
        "repr.url": "orange1",
    }
)

console = Console(theme=skore_console_theme, width=88)

joblib_version = parse_version(joblib.__version__)
if joblib_version < parse_version("1.4"):
    warnings.warn(
        "Because your version of joblib is older than 1.4, some of the features of "
        "skore will not be enabled (e.g. progress bars). You can update joblib to "
        "benefit from these features.",
        stacklevel=2,
    )
    set_config(show_progress=False)



from typing import Any, Protocol, runtime_checkable

import pandas as pd


@runtime_checkable
class Display(Protocol):
    """Protocol specifying the common API for all `skore` displays."""

    def plot(self, **kwargs: Any) -> None:
        """Display a figure containing the information of the display."""

    def set_style(self, **kwargs: Any) -> None:
        """Set the style of the display."""

    def frame(self, **kwargs: Any) -> pd.DataFrame:
        """Get the data used to create the display.

        Returns
        -------
        DataFrame
            A DataFrame containing the data used to create the display.
        """



import contextlib
import copy
from typing import Literal

import numpy as np
import pandas as pd
from matplotlib.legend import Legend
from sklearn.base import BaseEstimator, ClassifierMixin

from skore._sklearn._plot.metrics.precision_recall_curve import (
    PrecisionRecallCurveDisplay,
)
from skore._sklearn._plot.metrics.roc_curve import RocCurveDisplay


@contextlib.contextmanager
def check_cache_changed(value):
    """Assert that `value` has changed during context execution."""
    initial_value = copy.copy(value)
    yield
    assert value != initial_value


@contextlib.contextmanager
def check_cache_unchanged(value):
    """Assert that `value` has not changed during context execution."""
    initial_value = copy.copy(value)
    yield
    assert value == initial_value


class MockEstimator(ClassifierMixin, BaseEstimator):
    def __init__(self, *, error, n_call=0, fail_after_n_clone=3):
        self.error = error
        self.n_call = n_call
        self.fail_after_n_clone = fail_after_n_clone

    def fit(self, X, y):
        if self.n_call > self.fail_after_n_clone:
            raise self.error
        self.classes_ = np.unique(y)
        return self

    def __sklearn_clone__(self):
        self.n_call += 1
        return self

    def predict(self, X):
        return np.ones(X.shape[0])


def check_roc_curve_display_data(display: RocCurveDisplay):
    """Check the structure of the display's internal data."""
    assert list(display.roc_curve.columns) == [
        "estimator_name",
        "split_index",
        "label",
        "threshold",
        "fpr",
        "tpr",
    ]
    assert list(display.roc_auc.columns) == [
        "estimator_name",
        "split_index",
        "label",
        "roc_auc",
    ]


def check_precision_recall_curve_display_data(display: PrecisionRecallCurveDisplay):
    """Check the structure of the display's internal data."""
    assert list(display.precision_recall.columns) == [
        "estimator_name",
        "split_index",
        "label",
        "threshold",
        "precision",
        "recall",
    ]
    assert list(display.average_precision.columns) == [
        "estimator_name",
        "split_index",
        "label",
        "average_precision",
    ]


def check_legend_position(ax, *, loc: str, position: Literal["inside", "outside"]):
    """Check the position of the legend in the axes."""
    legend = ax.get_legend()
    assert legend._loc == Legend.codes[loc]
    bbox = legend.get_window_extent().transformed(ax.transAxes.inverted())
    if position == "inside":
        assert 0 <= bbox.x0 <= 1
    else:
        assert bbox.x0 >= 1


def check_frame_structure(df, expected_index, expected_data_columns):
    """Check the structure of ROC, precision-recall, and prediction error DataFrames.

    Parameters
    ----------
    df : DataFrame
        The DataFrame to check.
    expected_index : list of str
        The expected index column names (e.g., `estimator_name`, `split_index`,
        `label`).
        These columns should be of categorical type.
    expected_data_columns : list of str
        The expected data column names (e.g., `threshold`, `fpr`, `tpr`,
        `precision`, `recall`).
        These columns should be of `np.float64` type.
    """
    assert isinstance(df, pd.DataFrame)
    assert sorted(df.columns.tolist()) == sorted(expected_index + expected_data_columns)

    for col in df.columns:
        if col in expected_index:
            assert df[col].dtype.name == "category"
        else:
            assert col in expected_data_columns
            assert df[col].dtype == np.float64



from collections.abc import Sequence
from typing import Any, Literal, cast

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from numpy.typing import NDArray
from pandas import DataFrame
from sklearn.base import BaseEstimator
from sklearn.metrics import average_precision_score, precision_recall_curve
from sklearn.preprocessing import LabelBinarizer

from skore._sklearn._plot.style import StyleDisplayMixin
from skore._sklearn._plot.utils import (
    LINESTYLE,
    HelpDisplayMixin,
    PlotBackendMixin,
    _ClassifierCurveDisplayMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)
from skore._sklearn.types import MLTask, PositiveLabel, ReportType, YPlotData


def _set_axis_labels(ax: Axes, info_pos_label: str | None) -> None:
    """Add axis labels."""
    xlabel = "Recall"
    ylabel = "Precision"
    if info_pos_label:
        xlabel += info_pos_label
        ylabel += info_pos_label

    ax.set(
        xlabel=xlabel,
        xlim=(-0.01, 1.01),
        ylabel=ylabel,
        ylim=(-0.01, 1.01),
        aspect="equal",
    )


MAX_N_LABELS = 5


class PrecisionRecallCurveDisplay(
    StyleDisplayMixin, HelpDisplayMixin, _ClassifierCurveDisplayMixin, PlotBackendMixin
):
    """Precision Recall visualization.

    An instance of this class should be created by
    `EstimatorReport.metrics.precision_recall()`. You should not create an
    instance of this class directly.

    Parameters
    ----------
    precision_recall : DataFrame
        The precision-recall curve data to display. The columns are

        - `estimator_name`
        - `split_index` (may be null)
        - `label`
        - `threshold`
        - `precision`
        - `recall`.

    average_precision : DataFrame
        The average precision data to display. The columns are

        - `estimator_name`
        - `split_index` (may be null)
        - `label`
        - `average_precision`.

    pos_label : int, float, bool, str or None
        The class considered as the positive class. If None, the class will not
        be shown in the legend.

    data_source : {"train", "test", "X_y"}
        The data source used to compute the precision recall curve.

    ml_task : {"binary-classification", "multiclass-classification"}
        The machine learning task.

    report_type : {"comparison-cross-validation", "comparison-estimator", \
            "cross-validation", "estimator"}
        The type of report.

    Attributes
    ----------
    ax_ : matplotlib axes or ndarray of axes
        The axes on which the precision-recall curve is plotted.

    figure_ : matplotlib figure
        The figure on which the precision-recall curve is plotted.

    lines_ : list of matplotlib lines
        The lines of the precision-recall curve.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import train_test_split
    >>> from skore import EstimatorReport
    >>> X, y = load_breast_cancer(return_X_y=True)
    >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
    >>> classifier = LogisticRegression(max_iter=10_000)
    >>> report = EstimatorReport(classifier, **split_data)
    >>> display = report.metrics.precision_recall()
    >>> display.plot(pr_curve_kwargs={"color": "tab:red"})
    """

    _default_pr_curve_kwargs: dict[str, Any] | None = None

    def __init__(
        self,
        *,
        precision_recall: DataFrame,
        average_precision: DataFrame,
        pos_label: PositiveLabel | None,
        data_source: Literal["train", "test", "X_y"],
        ml_task: MLTask,
        report_type: ReportType,
    ) -> None:
        self.precision_recall = precision_recall
        self.average_precision = average_precision
        self.pos_label = pos_label
        self.data_source = data_source
        self.ml_task = ml_task
        self.report_type = report_type

    def _plot_single_estimator(
        self,
        *,
        estimator_name: str,
        pr_curve_kwargs: list[dict[str, Any]],
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot precision-recall curve for a single estimator.

        Parameters
        ----------
        estimator_name : str
            The name of the estimator.

        pr_curve_kwargs : list of dict
            Additional keyword arguments to pass to matplotlib's plot function. In
            binary case, we should have a single dict. In multiclass case, we should
            have a list of dicts, one per class.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes containing the plot.

        lines : list of matplotlib.lines.Line2D
            The plotted lines.

        info_pos_label : str or None
            String containing the positive label info for binary classification.
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {"drawstyle": "steps-post"}

        if self.ml_task == "binary-classification":
            precision_recall = self.precision_recall.query(
                f"label == {self.pos_label!r}"
            )
            average_precision = self.average_precision["average_precision"].item()

            line_kwargs_validated = _validate_style_kwargs(
                line_kwargs, pr_curve_kwargs[0]
            )
            if self.data_source in ("train", "test"):
                line_kwargs_validated["label"] = (
                    f"{self.data_source.title()} set (AP = {average_precision:0.2f})"
                )
            else:  # data_source in (None, "X_y")
                line_kwargs_validated["label"] = f"AP = {average_precision:0.2f}"

            (line,) = self.ax_.plot(
                precision_recall["recall"],
                precision_recall["precision"],
                **line_kwargs_validated,
            )
            lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
            legend_title = None

        else:  # multiclass-classification
            labels = self.precision_recall["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(labels) < 10 else len(labels),
            )

            for class_idx, class_label in enumerate(labels):
                query = f"label == {class_label!r}"
                precision_recall = self.precision_recall.query(query)
                average_precision = self.average_precision.query(query)[
                    "average_precision"
                ].item()
                pr_curve_kwargs_class = pr_curve_kwargs[class_idx]

                line_kwargs["color"] = class_colors[class_idx]
                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, pr_curve_kwargs_class
                )
                line_kwargs_validated["label"] = (
                    f"{str(class_label).title()} (AP = {average_precision:0.2f})"
                )

                (line,) = self.ax_.plot(
                    precision_recall["recall"],
                    precision_recall["precision"],
                    **line_kwargs_validated,
                )
                lines.append(line)

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = None
            info_pos_label = None  # irrelevant for multiclass

        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower left", title=legend_title)
        self.ax_.set_title(f"Precision-Recall Curve for {estimator_name}")

        return self.ax_, lines, info_pos_label

    def _plot_cross_validated_estimator(
        self,
        *,
        estimator_name: str,
        pr_curve_kwargs: list[dict[str, Any]],
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot precision-recall curve for a cross-validated estimator.

        Parameters
        ----------
        estimator_name : str
            The name of the estimator.

        pr_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the
            precision-recall curves. The length of the list should match the number of
            curves to plot.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the precision-recall curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted precision-recall curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {"drawstyle": "steps-post"}

        if self.ml_task == "binary-classification":
            for split_idx in self.precision_recall["split_index"].cat.categories:
                query = f"label == {self.pos_label!r} & split_index == {split_idx}"
                precision_recall = self.precision_recall.query(query)
                average_precision = self.average_precision.query(query)[
                    "average_precision"
                ].item()

                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, pr_curve_kwargs[split_idx]
                )
                line_kwargs_validated["label"] = (
                    f"Fold #{split_idx + 1} (AP = {average_precision:0.2f})"
                )

                (line,) = self.ax_.plot(
                    precision_recall["recall"],
                    precision_recall["precision"],
                    **line_kwargs_validated,
                )
                lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.precision_recall["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(labels) < 10 else len(labels),
            )

            for class_idx, class_label in enumerate(labels):
                # precision_class = self.precision[class_]
                # recall_class = self.recall[class_]
                # average_precision_class = self.average_precision[class_]
                pr_curve_kwargs_class = pr_curve_kwargs[class_idx]

                for split_idx in self.precision_recall["split_index"].cat.categories:
                    query = f"label == {class_label!r} & split_index == {split_idx}"
                    precision_recall = self.precision_recall.query(query)
                    average_precision = self.average_precision.query(query)[
                        "average_precision"
                    ]
                    average_precision_mean = np.mean(average_precision)
                    average_precision_std = np.std(average_precision, ddof=1)

                    line_kwargs["color"] = class_colors[class_idx]
                    line_kwargs["alpha"] = 0.3
                    line_kwargs_validated = _validate_style_kwargs(
                        line_kwargs, pr_curve_kwargs_class
                    )
                    if split_idx == 0:
                        line_kwargs_validated["label"] = (
                            f"{str(class_label).title()} "
                            f"(AP = {average_precision_mean:0.2f} +/- "
                            f"{average_precision_std:0.2f})"
                        )
                    else:
                        line_kwargs_validated["label"] = None

                    (line,) = self.ax_.plot(
                        precision_recall["recall"],
                        precision_recall["precision"],
                        **line_kwargs_validated,
                    )
                    lines.append(line)

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"
        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower left", title=legend_title)
        self.ax_.set_title(f"Precision-Recall Curve for {estimator_name}")

        return self.ax_, lines, info_pos_label

    def _plot_comparison_estimator(
        self,
        *,
        estimator_names: list[str],
        pr_curve_kwargs: list[dict[str, Any]],
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot precision-recall curve of several estimators.

        Parameters
        ----------
        estimator_names : list of str
            The names of the estimators.

        pr_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the
            precision-recall curves. The length of the list should match the number of
            curves to plot.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the precision-recall curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted precision-recall curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {"drawstyle": "steps-post"}

        if self.ml_task == "binary-classification":
            for est_idx, est_name in enumerate(estimator_names):
                query = f"label == {self.pos_label!r} & estimator_name == '{est_name}'"
                precision_recall = self.precision_recall.query(query)
                average_precision = self.average_precision.query(query)[
                    "average_precision"
                ].item()

                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, pr_curve_kwargs[est_idx]
                )
                line_kwargs_validated["label"] = (
                    f"{est_name} (AP = {average_precision:0.2f})"
                )
                (line,) = self.ax_.plot(
                    precision_recall["recall"],
                    precision_recall["precision"],
                    **line_kwargs_validated,
                )
                lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.precision_recall["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(labels) < 10 else len(labels),
            )

            for est_idx, est_name in enumerate(estimator_names):
                est_color = class_colors[est_idx]

                for class_idx, class_label in enumerate(labels):
                    query = f"label == {class_label!r} & estimator_name == '{est_name}'"
                    precision_recall = self.precision_recall.query(query)
                    average_precision = self.average_precision.query(query)[
                        "average_precision"
                    ].item()

                    class_linestyle = LINESTYLE[(class_idx % len(LINESTYLE))][1]
                    line_kwargs["color"] = est_color
                    line_kwargs["alpha"] = 0.6
                    line_kwargs["linestyle"] = class_linestyle
                    line_kwargs_validated = _validate_style_kwargs(
                        line_kwargs, pr_curve_kwargs[est_idx]
                    )
                    line_kwargs_validated["label"] = (
                        f"{est_name} - {str(class_label).title()} "
                        f"(AP = {average_precision:0.2f})"
                    )

                    (line,) = self.ax_.plot(
                        precision_recall["recall"],
                        precision_recall["precision"],
                        **line_kwargs_validated,
                    )
                    lines.append(line)

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower left", title=legend_title)
        self.ax_.set_title("Precision-Recall Curve")

        return self.ax_, lines, info_pos_label

    def _plot_comparison_cross_validation(
        self,
        *,
        estimator_names: list[str],
        pr_curve_kwargs: list[dict[str, Any]],
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot precision-recall curve of several estimators.

        Parameters
        ----------
        estimator_names : list of str
            The names of the estimators.

        pr_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the
            precision-recall curves. The length of the list should match the number of
            curves to plot.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the precision-recall curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted precision-recall curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {"drawstyle": "steps-post"}

        if self.ml_task == "binary-classification":
            labels = self.precision_recall["label"].cat.categories
            colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(estimator_names) < 10 else len(estimator_names),
            )
            curve_idx = 0
            for report_idx, estimator_name in enumerate(estimator_names):
                query = (
                    f"label == {self.pos_label!r} "
                    f"& estimator_name == '{estimator_name}'"
                )
                average_precision = self.average_precision.query(query)[
                    "average_precision"
                ]

                precision_recall = self.precision_recall.query(query)

                for split_idx, segment in precision_recall.groupby(
                    "split_index", observed=True
                ):
                    if split_idx == 0:
                        label_kwargs = {
                            "label": (
                                f"{estimator_name} "
                                f"(AUC = {average_precision.mean():0.2f} "
                                f"+/- {average_precision.std():0.2f})"
                            )
                        }
                    else:
                        label_kwargs = {}

                    line_kwargs["color"] = colors[report_idx]
                    line_kwargs["alpha"] = 0.6
                    line_kwargs_validated = _validate_style_kwargs(
                        line_kwargs, pr_curve_kwargs[curve_idx]
                    )

                    (line,) = self.ax_.plot(
                        segment["recall"],
                        segment["precision"],
                        **(line_kwargs_validated | label_kwargs),
                    )
                    lines.append(line)

                    curve_idx += 1

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = "External set"

            _, labels = self.ax_.get_legend_handles_labels()
            if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
                self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
            else:
                self.ax_.legend(loc="lower left", title=legend_title)
            self.ax_.set_title("Precision-Recall Curve")

        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.precision_recall["label"].cat.categories
            colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(estimator_names) < 10 else len(estimator_names),
            )
            idx = 0

            for est_idx, estimator_name in enumerate(estimator_names):
                est_color = colors[est_idx]

                for label_idx, label in enumerate(labels):
                    query = f"label == {label!r} & estimator_name == '{estimator_name}'"
                    average_precision = self.average_precision.query(query)[
                        "average_precision"
                    ]

                    precision_recall = self.precision_recall.query(query)

                    for split_idx, segment in precision_recall.groupby(
                        "split_index", observed=True
                    ):
                        if split_idx == 0:
                            label_kwargs = {
                                "label": (
                                    f"{estimator_name} "
                                    f"(AUC = {average_precision.mean():0.2f} "
                                    f"+/- {average_precision.std():0.2f})"
                                )
                            }
                        else:
                            label_kwargs = {}

                        line_kwargs["color"] = est_color
                        line_kwargs["alpha"] = 0.6
                        line_kwargs_validated = _validate_style_kwargs(
                            line_kwargs, pr_curve_kwargs[idx]
                        )

                        (line,) = self.ax_[label_idx].plot(
                            segment["recall"],
                            segment["precision"],
                            **(line_kwargs_validated | label_kwargs),
                        )
                        lines.append(line)

                        idx = idx + 1

                    info_pos_label = f"\n(Positive label: {label})"
                    _set_axis_labels(self.ax_[label_idx], info_pos_label)

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = "External set"

            for ax in self.ax_:
                _, labels = ax.get_legend_handles_labels()
                ax.legend(loc="lower left", title=legend_title)

            self.figure_.suptitle("Precision-Recall Curve")

        return self.ax_, lines, info_pos_label

    @StyleDisplayMixin.style_plot
    def _plot_matplotlib(
        self,
        *,
        estimator_name: str | None = None,
        pr_curve_kwargs: dict[str, Any] | list[dict[str, Any]] | None = None,
        despine: bool = True,
    ) -> None:
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's `plot`.

        Parameters
        ----------
        estimator_name : str, default=None
            Name of the estimator used to plot the precision-recall curve. If
            `None`, we use the inferred name from the estimator.

        pr_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the precision-recall curve(s).

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Notes
        -----
        The average precision (cf. :func:`~sklearn.metrics.average_precision_score`)
        in scikit-learn is computed without any interpolation. To be consistent
        with this metric, the precision-recall curve is plotted without any
        interpolation as well (step-wise style).

        You can change this style by passing the keyword argument
        `drawstyle="default"`. However, the curve will not be strictly
        consistent with the reported average precision.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> display = report.metrics.precision_recall()
        >>> display.plot(pr_curve_kwargs={"color": "tab:red"})
        """
        if (
            self.report_type == "comparison-cross-validation"
            and self.ml_task == "multiclass-classification"
        ):
            n_labels = len(self.average_precision["label"].cat.categories)
            self.figure_, self.ax_ = plt.subplots(
                ncols=n_labels, figsize=(6.4 * n_labels, 4.8)
            )
        else:
            self.figure_, self.ax_ = plt.subplots()

        if pr_curve_kwargs is None:
            pr_curve_kwargs = self._default_pr_curve_kwargs

        if self.ml_task == "binary-classification":
            n_curves = len(self.average_precision.query(f"label == {self.pos_label!r}"))
        else:
            n_curves = len(self.average_precision)

        pr_curve_kwargs = self._validate_curve_kwargs(
            curve_param_name="pr_curve_kwargs",
            curve_kwargs=pr_curve_kwargs,
            n_curves=n_curves,
            report_type=self.report_type,
        )

        if self.report_type == "estimator":
            self.ax_, self.lines_, info_pos_label = self._plot_single_estimator(
                estimator_name=(
                    self.precision_recall["estimator_name"].cat.categories.item()
                    if estimator_name is None
                    else estimator_name
                ),
                pr_curve_kwargs=pr_curve_kwargs,
            )
        elif self.report_type == "cross-validation":
            self.ax_, self.lines_, info_pos_label = (
                self._plot_cross_validated_estimator(
                    estimator_name=(
                        self.precision_recall["estimator_name"].cat.categories.item()
                        if estimator_name is None
                        else estimator_name
                    ),
                    pr_curve_kwargs=pr_curve_kwargs,
                )
            )
        elif self.report_type == "comparison-estimator":
            self.ax_, self.lines_, info_pos_label = self._plot_comparison_estimator(
                estimator_names=self.precision_recall["estimator_name"].cat.categories,
                pr_curve_kwargs=pr_curve_kwargs,
            )
        elif self.report_type == "comparison-cross-validation":
            self.ax_, self.lines_, info_pos_label = (
                self._plot_comparison_cross_validation(
                    estimator_names=self.precision_recall[
                        "estimator_name"
                    ].cat.categories,
                    pr_curve_kwargs=pr_curve_kwargs,
                )
            )
        else:
            raise ValueError(
                "`report_type` should be one of 'estimator', 'cross-validation', "
                "'comparison-cross-validation' or 'comparison-estimator'. "
                f"Got '{self.report_type}' instead."
            )

        if (
            self.report_type == "comparison-cross-validation"
            and self.ml_task == "multiclass-classification"
        ):
            for ax in self.ax_:
                if despine:
                    _despine_matplotlib_axis(ax)
        else:
            _set_axis_labels(self.ax_, info_pos_label)

            if despine:
                _despine_matplotlib_axis(self.ax_)

    @classmethod
    def _compute_data_for_display(
        cls,
        y_true: Sequence[YPlotData],
        y_pred: Sequence[YPlotData],
        *,
        report_type: ReportType,
        estimators: Sequence[BaseEstimator],
        ml_task: MLTask,
        data_source: Literal["train", "test", "X_y"],
        pos_label: PositiveLabel | None,
        drop_intermediate: bool = True,
    ) -> "PrecisionRecallCurveDisplay":
        """Plot precision-recall curve given binary class predictions.

        Parameters
        ----------
        y_true : list of array-like of shape (n_samples,)
            True binary labels.

        y_pred : list of array-like of shape (n_samples,)
            Target scores, can either be probability estimates of the positive class,
            confidence values, or non-thresholded measure of decisions (as returned by
            "decision_function" on some classifiers).

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        estimators : list of estimator instances
            The estimators from which `y_pred` is obtained.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}
            The data source used to compute the precision recall curve.

        pos_label : int, float, bool, str or none
            The class considered as the positive class when computing the
            precision and recall metrics.

        drop_intermediate : bool, default=True
            Whether to drop some suboptimal thresholds which would not appear
            on a plotted precision-recall curve. This is useful in order to
            create lighter precision-recall curves.

        Returns
        -------
        display : PrecisionRecallCurveDisplay
        """
        pos_label_validated = cls._validate_from_predictions_params(
            y_true, y_pred, ml_task=ml_task, pos_label=pos_label
        )

        precision_recall_records = []
        average_precision_records = []

        if ml_task == "binary-classification":
            for y_true_i, y_pred_i in zip(y_true, y_pred, strict=False):
                pos_label_validated = cast(PositiveLabel, pos_label_validated)
                precision_i, recall_i, thresholds_i = precision_recall_curve(
                    y_true_i.y,
                    y_pred_i.y,
                    pos_label=pos_label_validated,
                    drop_intermediate=drop_intermediate,
                )
                average_precision_i = average_precision_score(
                    y_true_i.y, y_pred_i.y, pos_label=pos_label_validated
                )

                for precision, recall, threshold in zip(
                    precision_i, recall_i, thresholds_i, strict=False
                ):
                    precision_recall_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "label": pos_label_validated,
                            "threshold": threshold,
                            "precision": precision,
                            "recall": recall,
                        }
                    )
                average_precision_records.append(
                    {
                        "estimator_name": y_true_i.estimator_name,
                        "split_index": y_true_i.split_index,
                        "label": pos_label_validated,
                        "average_precision": average_precision_i,
                    }
                )
        else:  # multiclass-classification
            for y_true_i, y_pred_i, est in zip(
                y_true, y_pred, estimators, strict=False
            ):
                label_binarizer = LabelBinarizer().fit(est.classes_)
                y_true_onehot_i: NDArray = label_binarizer.transform(y_true_i.y)
                for class_idx, class_ in enumerate(est.classes_):
                    precision_class_i, recall_class_i, thresholds_class_i = (
                        precision_recall_curve(
                            y_true_onehot_i[:, class_idx],
                            y_pred_i.y[:, class_idx],
                            pos_label=None,
                            drop_intermediate=drop_intermediate,
                        )
                    )
                    average_precision_class_i = average_precision_score(
                        y_true_onehot_i[:, class_idx], y_pred_i.y[:, class_idx]
                    )

                    for precision, recall, threshold in zip(
                        precision_class_i,
                        recall_class_i,
                        thresholds_class_i,
                        strict=False,
                    ):
                        precision_recall_records.append(
                            {
                                "estimator_name": y_true_i.estimator_name,
                                "split_index": y_true_i.split_index,
                                "label": class_,
                                "threshold": threshold,
                                "precision": precision,
                                "recall": recall,
                            }
                        )
                    average_precision_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "label": class_,
                            "average_precision": average_precision_class_i,
                        }
                    )

        dtypes = {
            "estimator_name": "category",
            "split_index": "category",
            "label": "category",
        }

        return cls(
            precision_recall=DataFrame.from_records(precision_recall_records).astype(
                dtypes
            ),
            average_precision=DataFrame.from_records(average_precision_records).astype(
                dtypes
            ),
            pos_label=pos_label_validated,
            data_source=data_source,
            ml_task=ml_task,
            report_type=report_type,
        )

    def frame(self, with_average_precision: bool = False) -> DataFrame:
        """Get the data used to create the precision-recall curve plot.

        Parameters
        ----------
        with_average_precision : bool, default=False
            Whether to include the average precision column in the returned DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the precision-recall curve data with columns
            depending on the report type:

            - `estimator_name`: Name of the estimator (when comparing estimators)
            - `split_index`: Cross-validation fold ID (when doing cross-validation)
            - `label`: Class label (for multiclass-classification)
            - `threshold`: Decision threshold
            - `precision`: Precision score at threshold
            - `recall`: Recall score at threshold
            - `average_precision`: average precision
              (when `with_average_precision=True`)

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split, EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> clf = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(clf, **split_data)
        >>> display = report.metrics.precision_recall()
        >>> df = display.frame()
        """
        if with_average_precision:
            # The merge between the precision-recall curve and the average precision is
            # done without specifying the columns to merge on, hence done on all column
            # that are present in both DataFrames.
            # In this case, the common columns are all columns but not the ones
            # containing the statistics.
            df = self.precision_recall.merge(self.average_precision)
        else:
            df = self.precision_recall

        statistical_columns = ["threshold", "precision", "recall"]
        if with_average_precision:
            statistical_columns.append("average_precision")

        if self.report_type == "estimator":
            indexing_columns = []
        elif self.report_type == "cross-validation":
            indexing_columns = ["split_index"]
        elif self.report_type == "comparison-estimator":
            indexing_columns = ["estimator_name"]
        else:  # self.report_type == "comparison-cross-validation"
            indexing_columns = ["estimator_name", "split_index"]

        if self.ml_task == "binary-classification":
            columns = indexing_columns + statistical_columns
        else:
            columns = indexing_columns + ["label"] + statistical_columns

        return df[columns]



"""Module to handle the verbosity of a given logger."""

import logging
from collections.abc import Iterator
from contextlib import contextmanager

from rich.logging import LogRecord, RichHandler, Text


class Handler(RichHandler):
    """A logging handler that renders output with Rich."""

    def get_level_text(self, record: LogRecord) -> Text:
        """Get the logger name and levelname from the record."""
        levelname = record.levelname
        name = record.name
        txt = f"<Logger {name} ({levelname})>"

        return Text.styled(
            txt.ljust(8),
            f"logging.level.{levelname.lower()}",
        )


@contextmanager
def logger_context(logger: logging.Logger, verbose: bool = False) -> Iterator[None]:
    """Context manager for temporarily adding a Rich handler to a logger."""
    handler = None
    try:
        if verbose:
            formatter = logging.Formatter("%(message)s")
            handler = Handler(markup=True)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        yield
    finally:
        if verbose and handler:
            logger.removeHandler(handler)



"""Various utilities to help with development."""

from base64 import b64decode, b64encode


def bytes_to_b64_str(literal: bytes) -> str:
    """Encode the bytes-like object `literal` in a Base64 str."""
    return b64encode(literal).decode("utf-8")


def b64_str_to_bytes(literal: str) -> bytes:
    """Decode the Base64 str object `literal` in a bytes."""
    return b64decode(literal.encode("utf-8"))


__all__ = [
    "bytes_to_b64_str",
    "b64_str_to_bytes",
]



from skore._externals._pandas_accessors import _register_accessor
from skore._sklearn._cross_validation.metrics_accessor import _MetricsAccessor
from skore._sklearn._cross_validation.report import (
    CrossValidationReport,
)

_register_accessor("metrics", CrossValidationReport)(_MetricsAccessor)

__all__ = ["CrossValidationReport"]



from typing import Any

import joblib

from skore._externals._sklearn_compat import parse_version


# FIXME: Remove this function once we support only joblib >= 1.4
def _validate_joblib_parallel_params(**kwargs: Any) -> dict[str, Any]:
    """Validate the parameters for `joblib.Parallel`.

    Currently this function is in charge of removing the parameter `return_as`
    because it is not supported by joblib < 1.4.
    """
    joblib_version = parse_version(joblib.__version__)
    if joblib_version < parse_version("1.4"):
        kwargs.pop("return_as", None)
    return kwargs



"""A helper to guess the machine-learn task being performed."""

import numpy as np
from sklearn.base import is_classifier, is_regressor
from sklearn.utils import check_array
from sklearn.utils.multiclass import type_of_target

from skore._externals._sklearn_compat import is_clusterer
from skore._sklearn.types import MLTask


def _column_is_classification(y) -> bool:
    """Check whether ``y`` is a sequence.

    We define a sequence as a 1-d array of sequential integer values,
    where the first value is 0.
    """
    y_unique = np.unique(y)
    if np.any(y_unique == 0):
        sequential = np.arange(y_unique[0], y_unique[-1] + 1)
        return np.array_equal(y_unique, sequential)
    return False


def _is_classification(y) -> bool:
    """Determine if `y` is a target for a classification task.

    If `y` contains integers, sklearn's `type_of_target` considers the task
    to be multiclass classification. This might not be the case, so we add the
    constraints that `y` must contain sequential values and contain 0.

    If `y` is a 2-d array, we check each column independently; if at least one
    column does not fit the constraints, then we consider the task to be regression.

    Note that this can cause false positives, e.g. a classification task with classes
    0, 1, 2 where `y` contains no examples of class 1 would be falsely considered
    "regression".
    Similarly, if `y` is a 2-d array where some column contains 0 and 2 but not 1,
    the whole array would be considered "regression".

    If `y` does not contain numbers (e.g. strings) then this function returns True.
    """
    try:
        y = check_array(y, dtype="numeric", ensure_2d=False)
    except ValueError:
        # The conversion in `check_array` failed meaning
        # that we are in a classification case with non-numeric
        # data type.
        return True

    if y.ndim == 1:
        return _column_is_classification(y)

    # Iterate on columns of y (check_array ensures that y is at most 2-d)
    return all(_column_is_classification(column) for column in y.T)


def _find_ml_task(y, estimator=None) -> MLTask:
    """Guess the ML task being addressed based on a target array and an estimator.

    This relies first on the estimator characteristics, and falls back on
    analyzing ``y``. Check the examples for some of the heuristics relied on.

    Parameters
    ----------
    y : numpy.ndarray
        A target vector.
    estimator : sklearn.base.BaseEstimator, optional
        An estimator, used mainly if fitted.

    Returns
    -------
    MLTask
        The guess of the kind of ML task being performed.

    Examples
    --------
    >>> import numpy
    >>> from skore._sklearn.find_ml_task import _find_ml_task

    # Discrete values, not sequential
    >>> _find_ml_task(numpy.array([1, 5, 9]))
    'regression'

    # Discrete values, not sequential, containing 0
    >>> _find_ml_task(numpy.array([0, 1, 5, 9]))
    'regression'

    # Discrete sequential values, containing 0
    >>> _find_ml_task(numpy.array([0, 1, 2]))
    'multiclass-classification'

    # Discrete sequential values, not containing 0
    >>> _find_ml_task(numpy.array([1, 3, 2]))
    'regression'

    # Discrete sequential values, containing 0, in a 2d array
    >>> _find_ml_task(numpy.array([[0, 1], [2, 2], [1, 0]]))
    'multioutput-multiclass-classification'

    # Discrete values, not sequential, in a 2d array
    >>> _find_ml_task(numpy.array([[1, 5], [5, 9]]))
    'multioutput-regression'

    # 2 columns, one of them not containing 0, in a 2d array
    >>> _find_ml_task(numpy.array([[0, 1], [1, 1]]))
    'multioutput-regression'

    # Discrete values, not sequential, containing 0, in a 2d array
    >>> _find_ml_task(numpy.array([[0, 1, 5, 9], [1, 0, 1, 1]]))
    'multioutput-regression'

    # 2 columns, one of them not sequential, in a 2d array
    >>> _find_ml_task(numpy.array([[0, 0], [2, 2], [1, 0]]))
    'multioutput-regression'
    """
    if estimator is not None:
        # checking the estimator is more robust and faster than checking the type of
        # target.
        if is_clusterer(estimator):
            return "clustering"
        if is_regressor(estimator):
            if y is None:
                return "regression"
            if np.ndim(y) == 1 or np.shape(y)[1] == 1:
                return "regression"
            return "multioutput-regression"
        if is_classifier(estimator):
            if hasattr(estimator, "classes_"):  # fitted estimator
                if (
                    isinstance(estimator.classes_, np.ndarray)
                    and estimator.classes_.ndim == 1
                ):
                    if estimator.classes_.size == 2:
                        return "binary-classification"
                    if estimator.classes_.size > 2:
                        return "multiclass-classification"
            else:
                # fallback on the target
                if y is None:
                    # we cannot do better than classification
                    return "classification"

    # fallback on the target
    if y is None:
        # NOTE: The task might not be clustering
        return "clustering"

    target_type = type_of_target(y)

    if target_type == "continuous":
        return "regression"
    if target_type == "continuous-multioutput":
        return "multioutput-regression"
    if target_type == "binary":
        return "binary-classification"
    if target_type == "multiclass":
        if _is_classification(y):
            return "multiclass-classification"
        return "regression"
    if target_type == "multiclass-multioutput":
        if _is_classification(y):
            return "multioutput-multiclass-classification"
        return "multioutput-regression"
    if target_type == "multilabel-indicator":
        if _is_classification(y):
            return "multioutput-binary-classification"
        return "multioutput-regression"
    return "unknown"



from skore._externals._pandas_accessors import _register_accessor
from skore._sklearn._comparison.metrics_accessor import _MetricsAccessor
from skore._sklearn._comparison.report import ComparisonReport

_register_accessor("metrics", ComparisonReport)(_MetricsAccessor)

__all__ = ["ComparisonReport"]



from skore._sklearn._plot.metrics.confusion_matrix import ConfusionMatrixDisplay
from skore._sklearn._plot.metrics.metrics_summary_display import MetricsSummaryDisplay
from skore._sklearn._plot.metrics.precision_recall_curve import (
    PrecisionRecallCurveDisplay,
)
from skore._sklearn._plot.metrics.prediction_error import PredictionErrorDisplay
from skore._sklearn._plot.metrics.roc_curve import RocCurveDisplay

__all__ = [
    "ConfusionMatrixDisplay",
    "PrecisionRecallCurveDisplay",
    "PredictionErrorDisplay",
    "RocCurveDisplay",
    "MetricsSummaryDisplay",
]



class ReprHTMLMixin:
    """Mixin to handle consistently the HTML representation.

    When inheriting from this class, you need to define an attribute `_html_repr`
    which is a callable that returns the HTML representation to be shown.
    """

    @property
    def _repr_html_(self):
        return self._repr_html_inner

    def _repr_html_inner(self):
        return self._html_repr()

    def _repr_mimebundle_(self, **kwargs):
        return {"text/plain": repr(self), "text/html": self._html_repr()}



"""'Shuffle is true' warning.

This warning is shown when ``shuffle`` is set to True.
"""

from __future__ import annotations

from skore._sklearn.train_test_split.warning.train_test_split_warning import (
    TrainTestSplitWarning,
)


class ShuffleTrueWarning(TrainTestSplitWarning):
    """Check whether ``shuffle`` is set to ``True``."""

    MSG = (
        "We detected that the `shuffle` parameter is set to `True` either explicitly "
        "or from its default value. In case of time-ordered events (even if they are "
        "independent), this will result in inflated model performance evaluation "
        "because natural drift will not be taken into account. We recommend setting "
        "the shuffle parameter to `False` in order to ensure the evaluation process is "
        "really representative of your production release process."
    )

    @staticmethod
    def check(
        shuffle: bool,
        **kwargs,
    ) -> str | None:
        """Check whether ``shuffle`` is set to ``True``.

        Parameters
        ----------
        shuffle : bool
            Whether to shuffle the data before splitting.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        if shuffle is not False:
            return ShuffleTrueWarning.MSG
        return None



import copy

import pandas as pd

from skore._sklearn.types import Aggregate


def _combine_estimator_results(
    individual_results: list[pd.DataFrame],
    estimator_names: list[str],
    indicator_favorability: bool,
) -> pd.DataFrame:
    """Combine a list of dataframes provided by `EstimatorReport`s.

    Parameters
    ----------
    results : pd.DataFrame
        The dataframes to combine.
        They are assumed to originate from a `EstimatorReport.metrics` computation.

    estimator_names : list of str of len (len(results))
        The name to give the estimator for each dataframe.

    indicator_favorability : bool
        Whether to keep the Favorability column.

    Examples
    --------
    >>> from skore._sklearn._comparison.utils import _combine_estimator_results
    >>> import pandas as pd
    >>> individual_results = [
    ...     pd.DataFrame.from_dict(
    ...         {
    ...             "index": ["Brier score"],
    ...             "columns": ["LogisticRegression"],
    ...             "data": [[0.026]],
    ...             "index_names": ["Metric"],
    ...             "column_names": [None],
    ...         },
    ...         orient="tight",
    ...     ),
    ...     pd.DataFrame.from_dict(
    ...         {
    ...             "index": ["Brier score"],
    ...             "columns": ["LogisticRegression"],
    ...             "data": [[0.026]],
    ...             "index_names": ["Metric"],
    ...             "column_names": [None],
    ...         },
    ...         orient="tight",
    ...     ),
    ... ]
    >>> estimator_names = ['LogisticRegression_1', 'LogisticRegression_2']
    >>> _combine_estimator_results(
    ...     individual_results,
    ...     estimator_names,
    ...     indicator_favorability=False,
    ... )
    Estimator    LogisticRegression_1  LogisticRegression_2
    Metric
    Brier score                   ...                   ...
    """
    results = pd.concat(individual_results, axis=1)

    # Pop the favorability column if it exists, to:
    # - not use it in the aggregate operation
    # - later to only report a single column and not by split columns
    if indicator_favorability:
        # Some metrics can be undefined for some estimators and NaN are
        # introduced after the concatenation. We fill the NaN using the
        # valid favorability
        favorability = results.pop("Favorability").bfill(axis=1).iloc[:, 0]
    else:
        favorability = None

    results.columns = pd.Index(estimator_names, name="Estimator")

    if favorability is not None:
        results["Favorability"] = favorability

    return results


def _combine_cross_validation_results(
    individual_results: list[pd.DataFrame],
    estimator_names: list[str],
    indicator_favorability: bool,
    aggregate: Aggregate | None,
) -> pd.DataFrame:
    """Combine a list of dataframes provided by `CrossValidationReport`s.

    Parameters
    ----------
    results : pd.DataFrame
        The dataframes to combine.
        They are assumed to originate from a `CrossValidationReport.metrics`
        computation. In particular, there are several assumptions made:

        - every dataframe has the form:
            - index: Index "Metric", or MultiIndex ["Metric", "Label / Average"]
            - columns: MultiIndex with levels ["Estimator", "Splits"] (can be
              unnamed)
        - all dataframes have the same metrics

        The dataframes are not required to have the same number of columns (splits).

    estimator_names : list of str of length `len(results)`
        The name to give the estimator for each dataframe.

    indicator_favorability : bool
        Whether to keep the Favorability column.

    aggregate : Aggregate
        How to aggregate the resulting dataframe.

    Examples
    --------
    >>> from skore._sklearn._comparison.utils import _combine_cross_validation_results
    >>> import pandas as pd
    >>> pd.set_option('display.max_columns', None)
    >>> pd.set_option('display.width', 1000)
    >>> individual_results = [
    ...     pd.DataFrame.from_dict(
    ...         {
    ...             "index": ["Accuracy"],
    ...             "columns": [
    ...                 ("DummyClassifier", "Split #0"),
    ...                 ("DummyClassifier", "Split #1"),
    ...                 ("DummyClassifier", "Split #2"),
    ...                 ("DummyClassifier", "Split #3"),
    ...                 ("DummyClassifier", "Split #4"),
    ...             ],
    ...             "data": [[0.45, 0.45, 0.35, 0.35, 0.55]],
    ...             "index_names": ["Metric"],
    ...             "column_names": [None, None],
    ...         },
    ...         orient="tight",
    ...     ),
    ...     pd.DataFrame.from_dict(
    ...         {
    ...             "index": ["Accuracy"],
    ...             "columns": [
    ...                 ("DummyClassifier", "Split #0"),
    ...                 ("DummyClassifier", "Split #1"),
    ...                 ("DummyClassifier", "Split #2"),
    ...             ],
    ...             "data": [[0.53, 0.42, 0.52]],
    ...             "index_names": ["Metric"],
    ...             "column_names": [None, None],
    ...         },
    ...         orient="tight",
    ...     ),
    ... ]
    >>> estimator_names = ["DummyClassifier_1", "DummyClassifier_2"]
    >>> _combine_cross_validation_results(
    ...     individual_results,
    ...     estimator_names,
    ...     indicator_favorability=False,
    ...     aggregate=None,
    ... )
                                            Value
    Metric   Estimator         Split
    Accuracy DummyClassifier_1 Split #0       ...
                               Split #1       ...
                               Split #2       ...
                               Split #3       ...
                               Split #4       ...
             DummyClassifier_2 Split #0       ...
                               Split #1       ...
                               Split #2       ...
    >>> _combine_cross_validation_results(
    ...     individual_results,
    ...     estimator_names,
    ...     indicator_favorability=False,
    ...     aggregate=["mean", "std"],
    ... )
                           mean                                 std
    Estimator DummyClassifier_1 DummyClassifier_2 DummyClassifier_1 DummyClassifier_2
    Metric
    Accuracy                ...               ...               ...               ...
    """

    def df_with_name(df: pd.DataFrame, estimator_name: str) -> pd.DataFrame:
        """Move the estimator name from the column index to the table index.

        Notes
        -----
        This mutates the input DataFrame.
        """
        df["Estimator"] = estimator_name

        if "Label / Average" in df.index.names:
            new_index = ["Metric", "Label / Average", "Estimator"]
        else:
            new_index = ["Metric", "Estimator"]
        df = df.reset_index().set_index(new_index)

        # Then drop the `estimator_name` from the columns index
        df.columns = df.columns.droplevel(0)

        return df

    def melt(df):
        """Move split number from columns to index."""
        df_reset = df.reset_index()

        if "Label / Average" in df_reset.columns:
            id_vars = ["Metric", "Label / Average", "Estimator"]
        else:
            id_vars = ["Metric", "Estimator"]

        melted = pd.melt(
            df_reset,
            id_vars=id_vars,
            var_name="Split",
            value_name="Value",
        )

        return melted

    def sort_by_split(df: pd.DataFrame) -> pd.DataFrame:
        """Sort dataframe, preserving the order of metrics.

        Ensures that the Split column goes [0, 1, 2, 0, 1, 2] rather than
        [0, 0, 1, 1, 2, 2].

        Notes
        -----
        Mutates the input dataframe.

        Examples
        --------
        >>> df = pd.DataFrame.from_dict(
        ...     {
        ...         "index": range(8),
        ...         "columns": [
        ...             "Metric",
        ...             "Label / Average",
        ...             "Estimator",
        ...             "Split",
        ...             "Value",
        ...         ],
        ...         "data": [
        ...             ["Precision", 0, "DummyClassifier_1", "Split #0", 0.44],
        ...             ["Precision", 1, "DummyClassifier_1", "Split #0", 0.45],
        ...             ["Precision", 0, "DummyClassifier_1", "Split #1", 0.44],
        ...             ["Precision", 1, "DummyClassifier_1", "Split #1", 0.45],
        ...             ["Precision", 0, "DummyClassifier_2", "Split #0", 0.53],
        ...             ["Precision", 1, "DummyClassifier_2", "Split #0", 0.52],
        ...             ["Precision", 0, "DummyClassifier_2", "Split #1", 0.42],
        ...             ["Precision", 1, "DummyClassifier_2", "Split #1", 0.42],
        ...         ],
        ...         "index_names": [None],
        ...         "column_names": [None],
        ...     },
        ...     orient="tight",
        ... )
        >>> sort_by_split(df)
              Metric  Label / Average          Estimator     Split  Value
        0  Precision                0  DummyClassifier_1  Split #0   0.44
        1  Precision                0  DummyClassifier_1  Split #1   0.44
        2  Precision                0  DummyClassifier_2  Split #0   0.53
        3  Precision                0  DummyClassifier_2  Split #1   0.42
        4  Precision                1  DummyClassifier_1  Split #0   0.45
        5  Precision                1  DummyClassifier_1  Split #1   0.45
        6  Precision                1  DummyClassifier_2  Split #0   0.52
        7  Precision                1  DummyClassifier_2  Split #1   0.42
        """
        df["Metric"] = df["Metric"].astype(
            pd.CategoricalDtype(df["Metric"].unique(), ordered=True)
        )
        df["metric_order_index"] = df["Metric"].cat.codes

        if "Label / Average" in df.columns:
            by = ["metric_order_index", "Label / Average", "Estimator", "Split"]
        else:
            by = ["metric_order_index", "Estimator", "Split"]

        df = (
            df.sort_values(by=by)
            .drop("metric_order_index", axis=1)
            .reset_index(drop=True)
        )

        df["Metric"] = df["Metric"].astype(str)

        return df

    # Deepcopy the contained dataframes to avoid mutating them;
    # in particular, they might be provided by a report cache
    results = copy.deepcopy(individual_results)

    # Pop the favorability column if it exists, to:
    # - not use it in the aggregate operation
    # - later to only report a single column and not by split columns
    if indicator_favorability:
        # Some metrics can be undefined for some estimators and NaN are
        # introduced after the concatenation. We fill the NaN using the
        # valid favorability
        favorability = (
            pd.concat([result.pop("Favorability") for result in results], axis=1)
            .bfill(axis=1)
            .iloc[:, 0]
        )
    else:
        favorability = None

    df = pd.concat(
        [
            melt(df_with_name(df, estimator_name))
            for df, estimator_name in zip(results, estimator_names, strict=False)
        ],
        axis=0,
    )

    if aggregate:
        if isinstance(aggregate, str):
            aggregate = [aggregate]

        if "Label / Average" in df.columns:
            index = ["Metric", "Label / Average"]
        else:
            index = ["Metric"]

        df = df.pivot_table(
            index=index,
            columns="Estimator",
            values="Value",
            aggfunc=aggregate,
            sort=False,
        )
    else:
        df = sort_by_split(df)
        df = df.set_index(list(df.columns.drop("Value")))

    if favorability is not None:
        df["Favorability"] = favorability

    return df



from skore._externals._pandas_accessors import _register_accessor
from skore._sklearn._estimator.data_accessor import _DataAccessor
from skore._sklearn._estimator.feature_importance_accessor import (
    _FeatureImportanceAccessor,
)
from skore._sklearn._estimator.metrics_accessor import _MetricsAccessor
from skore._sklearn._estimator.report import EstimatorReport

_register_accessor("metrics", EstimatorReport)(_MetricsAccessor)

_register_accessor("feature_importance", EstimatorReport)(_FeatureImportanceAccessor)

_register_accessor("data", EstimatorReport)(_DataAccessor)

__all__ = ["EstimatorReport"]



from __future__ import annotations

import copy
import time
import warnings
from itertools import product
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator, clone
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from skore._externals._pandas_accessors import DirNamesMixin
from skore._externals._sklearn_compat import is_clusterer
from skore._sklearn._base import _BaseReport, _get_cached_response_values
from skore._sklearn.find_ml_task import _find_ml_task
from skore._sklearn.types import _DEFAULT, PositiveLabel
from skore._utils._fixes import _validate_joblib_parallel_params
from skore._utils._measure_time import MeasureTime
from skore._utils._parallel import Parallel, delayed
from skore._utils._progress_bar import progress_decorator

if TYPE_CHECKING:
    from skore._sklearn._estimator.data_accessor import _DataAccessor
    from skore._sklearn._estimator.feature_importance_accessor import (
        _FeatureImportanceAccessor,
    )
    from skore._sklearn._estimator.metrics_accessor import _MetricsAccessor


class EstimatorReport(_BaseReport, DirNamesMixin):
    """Report for a fitted estimator.

    This class provides a set of tools to quickly validate and inspect a scikit-learn
    compatible estimator.

    Refer to the :ref:`estimator_report` section of the user guide for more details.

    Parameters
    ----------
    estimator : estimator object
        Estimator to make the report from. When the estimator is not fitted,
        it is deep-copied to avoid side-effects. If it is fitted, it is cloned instead.

    fit : {"auto", True, False}, default="auto"
        Whether to fit the estimator on the training data. If "auto", the estimator
        is fitted only if the training data is provided.

    X_train : {array-like, sparse matrix} of shape (n_samples, n_features) or \
            None
        Training data.

    y_train : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        Training target.

    X_test : {array-like, sparse matrix} of shape (n_samples, n_features) or None
        Testing data. It should have the same structure as the training data.

    y_test : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        Testing target.

    pos_label : int, float, bool or str, default=None
        For binary classification, the positive class. If `None` and the target labels
        are `{0, 1}` or `{-1, 1}`, the positive class is set to `1`. For other labels,
        some metrics might raise an error if `pos_label` is not defined.

    Attributes
    ----------
    estimator_ : estimator object
        The cloned or copied estimator.

    estimator_name_ : str
        The name of the estimator.

    fit_time_ : float or None
        The time taken to fit the estimator, in seconds. If the estimator is not
        internally fitted, the value is `None`.

    See Also
    --------
    skore.CrossValidationReport
        Report of cross-validation results.

    skore.ComparisonReport
        Report of comparison between estimators.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from skore import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> X, y = make_classification(random_state=42)
    >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
    >>> estimator = LogisticRegression()
    >>> from skore import EstimatorReport
    >>> report = EstimatorReport(estimator, **split_data)
    """

    _ACCESSOR_CONFIG: dict[str, dict[str, str]] = {
        "metrics": {"name": "metrics"},
        "feature_importance": {"name": "feature_importance"},
        "data": {"name": "data"},
    }

    metrics: _MetricsAccessor
    feature_importance: _FeatureImportanceAccessor
    data: _DataAccessor

    @staticmethod
    def _fit_estimator(
        estimator: BaseEstimator,
        X_train: ArrayLike | None,
        y_train: ArrayLike | None,
    ) -> tuple[BaseEstimator, float]:
        if X_train is None or (y_train is None and not is_clusterer(estimator)):
            raise ValueError(
                "The training data is required to fit the estimator. "
                "Please provide both X_train and y_train."
            )
        estimator_ = clone(estimator)
        with MeasureTime() as fit_time:
            estimator_.fit(X_train, y_train)
        return estimator_, fit_time()

    @classmethod
    def _copy_estimator(cls, estimator: BaseEstimator) -> BaseEstimator:
        try:
            return copy.deepcopy(estimator)
        except Exception as e:
            warnings.warn(
                "Deepcopy failed; using estimator as-is. "
                "Be aware that modifying the estimator outside of "
                f"{cls.__name__} will modify the internal estimator. "
                "Consider using a FrozenEstimator from scikit-learn to prevent this. "
                f"Original error: {e}",
                stacklevel=1,
            )
        return estimator

    def __init__(
        self,
        estimator: BaseEstimator,
        *,
        fit: Literal["auto"] | bool = "auto",
        X_train: ArrayLike | None = None,
        y_train: ArrayLike | None = None,
        X_test: ArrayLike | None = None,
        y_test: ArrayLike | None = None,
        pos_label: PositiveLabel | None = None,
    ) -> None:
        # used to know if a parent launch a progress bar manager
        self._progress_info: dict[str, Any] | None = None
        self._fit = fit

        fit_time: float | None = None
        if fit == "auto":
            try:
                check_is_fitted(estimator)
                self._estimator = self._copy_estimator(estimator)
            except NotFittedError:
                self._estimator, fit_time = self._fit_estimator(
                    estimator, X_train, y_train
                )
        elif fit is True:
            self._estimator, fit_time = self._fit_estimator(estimator, X_train, y_train)
        else:  # fit is False
            self._estimator = self._copy_estimator(estimator)

        # private storage to be able to invalidate the cache when the user alters
        # those attributes
        self._X_train = X_train
        self._y_train = y_train
        self._X_test = X_test
        self._y_test = y_test
        self._pos_label = pos_label
        self.fit_time_ = fit_time

        self._initialize_state()

    def _initialize_state(self) -> None:
        """Initialize/reset the random number generator, hash, and cache."""
        self._rng = np.random.default_rng(time.time_ns())
        self._hash = self._rng.integers(
            low=np.iinfo(np.int64).min, high=np.iinfo(np.int64).max
        )
        self._cache: dict[tuple[Any, ...], Any] = {}
        self._ml_task = _find_ml_task(self._y_test, estimator=self._estimator)

    # NOTE:
    # For the moment, we do not allow to alter the estimator and the training data.
    # For the validation set, we allow it and we invalidate the cache.

    def clear_cache(self) -> None:
        """Clear the cache.

        Note that the cache might not be empty after this method is run as some
        values need to be kept, such as the fit time.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> report.cache_predictions()
        >>> report.clear_cache()
        >>> report._cache
        {}
        """
        self._cache = {}

    @progress_decorator(description="Caching predictions")
    def cache_predictions(
        self,
        response_methods: Literal["auto"] | list[str] = "auto",
        n_jobs: int | None = None,
    ) -> None:
        """Cache estimator's predictions.

        Parameters
        ----------
        response_methods : "auto" or list of str, default="auto"
            The response methods to precompute. If "auto", the response methods are
            inferred from the ml task: for classification we compute the response of
            the `predict_proba`, `decision_function` and `predict` methods; for
            regression we compute the response of the `predict` method.

        n_jobs : int or None, default=None
            The number of jobs to run in parallel. None means 1 unless in a
            joblib.parallel_backend context. -1 means using all processors.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> report.cache_predictions()
        >>> report._cache
        {...}
        """
        if self._ml_task in ("binary-classification", "multiclass-classification"):
            if response_methods == "auto":
                response_methods = ["predict"]
                if hasattr(self._estimator, "predict_proba"):
                    response_methods += ["predict_proba"]
                if hasattr(self._estimator, "decision_function"):
                    response_methods += ["decision_function"]
            pos_labels = self._estimator.classes_.tolist() + [None]
        else:
            if response_methods == "auto":
                response_methods = ["predict"]
            pos_labels = [None]

        data_sources = [("test", self._X_test)]
        if self._X_train is not None:
            data_sources += [("train", self._X_train)]

        parallel = Parallel(
            **_validate_joblib_parallel_params(n_jobs=n_jobs, return_as="generator")
        )
        generator = parallel(
            delayed(_get_cached_response_values)(
                cache=self._cache,
                estimator_hash=self._hash,
                estimator=self._estimator,
                X=X,
                response_method=response_method,
                pos_label=pos_label,
                data_source=data_source,
            )
            for response_method, pos_label, (data_source, X) in product(
                response_methods, pos_labels, data_sources
            )
        )
        # trigger the computation
        assert self._progress_info is not None, (
            "The rich Progress class was not initialized."
        )
        progress = self._progress_info["current_progress"]
        task = self._progress_info["current_task"]
        total_iterations = len(response_methods) * len(pos_labels) * len(data_sources)
        progress.update(task, total=total_iterations)

        # do not mutate directly `self._cache` during the execution of Parallel
        results_to_cache = {}
        for results in generator:
            for key, value, is_cached in results:
                if not is_cached:
                    results_to_cache[key] = value
            progress.update(task, advance=1, refresh=True)

        if results_to_cache:
            self._cache.update(results_to_cache)

    def get_predictions(
        self,
        *,
        data_source: Literal["train", "test", "X_y"],
        response_method: Literal[
            "predict", "predict_proba", "decision_function"
        ] = "predict",
        X: ArrayLike | None = None,
        pos_label: PositiveLabel | None = _DEFAULT,
    ) -> ArrayLike:
        """Get estimator's predictions.

        This method has the advantage to reload from the cache if the predictions
        were already computed in a previous call.

        Parameters
        ----------
        data_source : {"test", "train", "X_y"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.
            - "X_y" : use the provided `X` and `y` to compute the predictions.

        response_method : {"predict", "predict_proba", "decision_function"}, \
                default="predict"
            The response method to use to get the predictions.

        X : array-like of shape (n_samples, n_features), optional
            When `data_source` is "X_y", the input features on which to compute the
            response method.

        pos_label : int, float, bool, str or None, default=_DEFAULT
            The label to consider as the positive class when computing predictions in
            binary classification cases. By default, the positive class is set to the
            one provided when creating the report. If `None`, `estimator_.classes_[1]`
            is used as positive label.

            When `pos_label` is equal to `estimator_.classes_[0]`, it will be equivalent
            to `estimator_.predict_proba(X)[:, 0]` for `response_method="predict_proba"`
            and `-estimator_.decision_function(X)` for
            `response_method="decision_function"`.

        Returns
        -------
        np.ndarray of shape (n_samples,) or (n_samples, n_classes)
            The predictions.

        Raises
        ------
        ValueError
            If the data source is invalid.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from skore import train_test_split
        >>> from sklearn.linear_model import LogisticRegression
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator = LogisticRegression()
        >>> from skore import EstimatorReport
        >>> report = EstimatorReport(estimator, **split_data)
        >>> predictions = report.get_predictions(data_source="test")
        >>> predictions.shape
        (25,)
        """
        if pos_label is _DEFAULT:
            pos_label = self.pos_label

        if data_source == "test":
            X_ = self._X_test
        elif data_source == "train":
            X_ = self._X_train
        elif data_source == "X_y":
            if X is None:
                raise ValueError(
                    "The `X` parameter is required when `data_source` is 'X_y'."
                )
            X_ = X
        else:
            raise ValueError(f"Invalid data source: {data_source}")

        results = _get_cached_response_values(
            cache=self._cache,
            estimator_hash=self._hash,
            estimator=self._estimator,
            X=X_,
            response_method=response_method,
            pos_label=pos_label,
            data_source=data_source,
        )
        for key, value, is_cached in results:
            if not is_cached:
                self._cache[key] = value
        return results[0][1]  # return the predictions only

    @property
    def ml_task(self) -> str:
        return self._ml_task

    @property
    def estimator(self) -> BaseEstimator:
        return self._estimator

    @property
    def estimator_(self) -> BaseEstimator:
        return self._estimator

    @property
    def X_train(self) -> ArrayLike | None:
        return self._X_train

    @property
    def y_train(self) -> ArrayLike | None:
        return self._y_train

    @property
    def X_test(self) -> ArrayLike | None:
        return self._X_test

    @X_test.setter
    def X_test(self, value):
        self._X_test = value
        self._initialize_state()

    @property
    def y_test(self) -> ArrayLike | None:
        return self._y_test

    @y_test.setter
    def y_test(self, value):
        self._y_test = value
        self._initialize_state()

    @property
    def pos_label(self) -> PositiveLabel | None:
        return self._pos_label

    @pos_label.setter
    def pos_label(self, value: PositiveLabel | None) -> None:
        self._pos_label = value
        self._initialize_state()

    @property
    def estimator_name_(self) -> str:
        if isinstance(self._estimator, Pipeline):
            name = self._estimator[-1].__class__.__name__
        else:
            name = self._estimator.__class__.__name__
        return name

    @property
    def fit(self) -> Literal["auto"] | bool:
        return self._fit

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_panel_title(self) -> str:
        return (
            f"[bold cyan]Tools to diagnose estimator {self.estimator_name_}[/bold cyan]"
        )

    def _get_help_legend(self) -> str:
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self) -> str:
        """Return a string representation."""
        return f"{self.__class__.__name__}(estimator={self.estimator_}, ...)"



import inspect
from collections.abc import Sequence
from io import StringIO
from typing import Any

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from rich.console import Console
from rich.panel import Panel
from rich.tree import Tree
from sklearn.utils.validation import (
    _check_pos_label_consistency,
    check_consistent_length,
)

from skore._config import get_config
from skore._sklearn.types import (
    MLTask,
    PlotBackend,
    PositiveLabel,
    ReportType,
    YPlotData,
)

LINESTYLE = [
    ("solid", "solid"),
    ("dotted", "dotted"),
    ("dashed", "dashed"),
    ("dashdot", "dashdot"),
    ("loosely dotted", (0, (1, 10))),
    ("dotted", (0, (1, 5))),
    ("densely dotted", (0, (1, 1))),
    ("long dash with offset", (5, (10, 3))),
    ("loosely dashed", (0, (5, 10))),
    ("dashed", (0, (5, 5))),
    ("densely dashed", (0, (5, 1))),
    ("loosely dashdotted", (0, (3, 10, 1, 10))),
    ("dashdotted", (0, (3, 5, 1, 5))),
    ("densely dashdotted", (0, (3, 1, 1, 1))),
    ("dashdotdotted", (0, (3, 5, 1, 5, 1, 5))),
    ("loosely dashdotdotted", (0, (3, 10, 1, 10, 1, 10))),
    ("densely dashdotdotted", (0, (3, 1, 1, 1, 1, 1))),
]


class HelpDisplayMixin:
    """Mixin class to add help functionality to a class."""

    estimator_name: str  # defined in the concrete display class

    def _get_attributes_for_help(self) -> list[str]:
        """Get the attributes ending with '_' to display in help."""
        attributes = []
        for name in dir(self):
            if name.endswith("_") and not name.startswith("_"):
                attributes.append(f".{name}")
        return sorted(attributes)

    def _get_methods_for_help(self) -> list[tuple[str, Any]]:
        """Get the public methods to display in help."""
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        filtered_methods = []
        for name, method in methods:
            is_private = name.startswith("_")
            is_class_method = inspect.ismethod(method) and method.__self__ is type(self)
            is_help_method = name == "help"
            if not (is_private or is_class_method or is_help_method):
                filtered_methods.append((f".{name}(...)", method))
        return sorted(filtered_methods)

    def _create_help_tree(self) -> Tree:
        """Create a rich Tree with attributes and methods."""
        tree = Tree("display")

        attributes = self._get_attributes_for_help()
        attr_branch = tree.add("[bold cyan] Attributes[/bold cyan]")
        # Ensure figure_ and ax_ are first
        sorted_attrs = sorted(attributes)
        if ("figure_" in sorted_attrs) and ("ax_" in sorted_attrs):
            sorted_attrs.remove(".ax_")
            sorted_attrs.remove(".figure_")
            sorted_attrs = [".figure_", ".ax_"] + [
                attr for attr in sorted_attrs if attr not in [".figure_", ".ax_"]
            ]
        for attr in sorted_attrs:
            attr_branch.add(attr)

        methods = self._get_methods_for_help()
        method_branch = tree.add("[bold cyan]Methods[/bold cyan]")
        for name, method in methods:
            description = (
                method.__doc__.split("\n")[0]
                if method.__doc__
                else "No description available"
            )
            method_branch.add(f"{name} - {description}")

        return tree

    def _create_help_panel(self) -> Panel:
        return Panel(
            self._create_help_tree(),
            title=f"[bold cyan]{self.__class__.__name__} [/bold cyan]",
            border_style="orange1",
            expand=False,
        )

    def help(self) -> None:
        """Display available attributes and methods using rich."""
        from skore import console  # avoid circular import

        console.print(self._create_help_panel())

    def __str__(self) -> str:
        """Return a string representation using rich."""
        string_buffer = StringIO()
        console = Console(file=string_buffer, force_terminal=False)
        console.print(
            Panel(
                "Get guidance using the help() method",
                title=f"[cyan]{self.__class__.__name__}[/cyan]",
                border_style="orange1",
                expand=False,
            )
        )
        return string_buffer.getvalue()

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        string_buffer = StringIO()
        console = Console(file=string_buffer, force_terminal=False)
        console.print(f"[cyan]skore.{self.__class__.__name__}(...)[/cyan]")
        return string_buffer.getvalue()


class _ClassifierCurveDisplayMixin:
    """Mixin class to be used in Displays requiring a binary classifier.

    The aim of this class is to centralize some validations regarding the estimator and
    the target and gather the response of the estimator.
    """

    # defined in the concrete display class
    estimator_name: str
    ml_task: MLTask
    pos_label: PositiveLabel | None

    def _validate_curve_kwargs(
        self,
        *,
        curve_param_name: str,
        curve_kwargs: dict[str, Any] | list[dict[str, Any]] | None,
        n_curves: int,
        report_type: ReportType,
    ) -> list[dict[str, Any]]:
        """Validate and format the classification curve keyword arguments.

        Parameters
        ----------
        curve_param_name : str
            The name of the curve parameter.

        curve_kwargs : dict or list of dict or None
            Keyword arguments to customize the classification curve.

        n_curves : int
            The number of curves we are plotting.

        report_type : {"comparison-cross-validation", "comparison-estimator",
                      "cross-validation", "estimator"}
            The type of report.

        Returns
        -------
        list of dict
            Validated list of keyword arguments for each curve.

        Raises
        ------
        ValueError
            If the format of curve_kwargs is invalid.
        """
        if self.ml_task == "binary-classification":
            if report_type in ("estimator", "cross-validation"):
                allow_single_dict = True
            elif report_type in ("comparison-estimator", "comparison-cross-validation"):
                # since we compare different estimators, it does not make sense to share
                # a single dictionary for all the estimators.
                allow_single_dict = False
            else:
                raise ValueError(
                    f"`report_type` should be one of 'estimator', 'cross-validation', "
                    "'comparison-cross-validation' or 'comparison-estimator'. "
                    f"Got '{report_type}' instead."
                )
        else:
            allow_single_dict = False

        if curve_kwargs is None:
            return [{}] * n_curves
        elif n_curves == 1:
            if isinstance(curve_kwargs, dict):
                return [curve_kwargs]
            elif isinstance(curve_kwargs, list) and len(curve_kwargs) == 1:
                return curve_kwargs
            else:
                raise ValueError(
                    "You intend to plot a single curve. We expect "
                    f"`{curve_param_name}` to be a dictionary. Got "
                    f"{type(curve_kwargs)} instead."
                )
        else:  # n_curves > 1
            if not allow_single_dict and isinstance(curve_kwargs, dict):
                raise ValueError(
                    "You intend to plot multiple curves. We expect "
                    f"`{curve_param_name}` to be a list of dictionaries. Got "
                    f"{type(curve_kwargs)} instead."
                )
            if isinstance(curve_kwargs, dict):
                return [curve_kwargs] * n_curves

            # From this point, we have a list of dictionaries
            if len(curve_kwargs) != n_curves:
                raise ValueError(
                    "You intend to plot multiple curves. We expect "
                    f"`{curve_param_name}` to be a list of dictionaries with the "
                    f"same length as the number of curves. Got "
                    f"{len(curve_kwargs)} instead of {n_curves}."
                )

            return curve_kwargs

    @classmethod
    def _validate_from_predictions_params(
        cls,
        y_true: Sequence[YPlotData],
        y_pred: Sequence[YPlotData],
        *,
        ml_task: str,
        pos_label: PositiveLabel | None = None,
    ) -> PositiveLabel | None:
        for y_true_i, y_pred_i in zip(y_true, y_pred, strict=False):
            check_consistent_length(y_true_i.y, y_pred_i.y)

        if ml_task == "binary-classification":
            pos_label = _check_pos_label_consistency(pos_label, y_true[0].y)

        return pos_label


def _rotate_ticklabels(
    ax: Axes, *, rotation: float = 45, horizontalalignment: str = "right"
) -> None:
    """Rotate the tick labels of the axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axis to rotate.
    rotation : float, default=45
        The rotation of the tick labels.
    horizontalalignment : {"left", "center", "right"}, default="right"
        The horizontal alignment of the tick labels.
    """
    plt.setp(
        ax.get_xticklabels(), rotation=rotation, horizontalalignment=horizontalalignment
    )


def _get_adjusted_fig_size(
    fig: Figure, ax: Axes, direction: str, target_size: float
) -> float:
    """Get the adjusted figure size.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure to adjust.

    ax : matplotlib.axes.Axes
        The axis to adjust.

    direction : {"width", "height"}
        The direction to adjust.

    target_size : float
        The target size.

    Returns
    -------
    float
        The adjusted figure size.
    """
    size_display = getattr(ax.get_window_extent(), direction)
    size = fig.dpi_scale_trans.inverted().transform((size_display, 0))[0]
    dim = 0 if direction == "width" else 1
    fig_size = fig.get_size_inches()[dim]
    return target_size * (fig_size / size)


def _adjust_fig_size(
    fig: Figure, ax: Axes, target_width: float, target_height: float
) -> None:
    """Rescale a figure to the target width and height.

    This allows us to have all figures of a given type (bar plots, lines or
    histograms) have the same size, so that the displayed report looks more
    uniform, without having to do manual adjustments to account for the length
    of labels, occurrence of titles etc. We let pyplot generate the figure
    without any size constraints then resize it and thus let matplotlib deal
    with resizing the appropriate elements (eg shorter bars when the labels
    take more horizontal space).

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure to adjust.

    ax : matplotlib.axes.Axes
        The axis to adjust.

    target_width : float
        The target width.

    target_height : float
        The target height.
    """
    width = _get_adjusted_fig_size(fig, ax, "width", target_width)
    height = _get_adjusted_fig_size(fig, ax, "height", target_height)
    fig.set_size_inches((width, height))


class PlotBackendMixin:
    """Mixin class for Displays to dispatch plotting to the configured backend."""

    def plot(self, **kwargs):
        plot_backend = get_config()["plot_backend"]
        if plot_backend == "matplotlib":
            return self._plot_matplotlib(**kwargs)
        elif plot_backend == "plotly":
            return self._plot_plotly(**kwargs)
        else:
            raise NotImplementedError(
                f"Plotting backend {plot_backend} not available. "
                f"Available options are {PlotBackend.__args__}."
            )

    def _plot_plotly(self, **kwargs):
        raise NotImplementedError(
            "Plotting with plotly is not supported for this Display."
        )


def _despine_matplotlib_axis(
    ax: Axes,
    *,
    axis_to_despine: tuple = ("top", "right"),
    x_range: tuple[float, float] = (0, 1),
    y_range: tuple[float, float] = (0, 1),
    offset: float = 10,
) -> None:
    """Despine the matplotlib axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axis to despine.
    x_range : tuple of float, default=(0, 1)
        The range of the x-axis.
    y_range : tuple of float, default=(0, 1)
        The range of the y-axis.
    offset : float, default=10
        The offset to add to the x-axis and y-axis.
    """
    ax.spines["bottom"].set_bounds(x_range[0], x_range[1])
    ax.spines["left"].set_bounds(y_range[0], y_range[1])
    ax.spines["left"].set_position(("outward", offset))
    ax.spines["bottom"].set_position(("outward", offset))
    for s in axis_to_despine:
        ax.spines[s].set_visible(False)


def _validate_style_kwargs(
    default_style_kwargs: dict[str, Any], user_style_kwargs: dict[str, Any]
) -> dict[str, Any]:
    """Create valid style kwargs by avoiding Matplotlib alias errors.

    Matplotlib raises an error when, for example, 'color' and 'c', or 'linestyle' and
    'ls', are specified together. To avoid this, we automatically keep only the one
    specified by the user and raise an error if the user specifies both.

    Parameters
    ----------
    default_style_kwargs : dict
        The Matplotlib style kwargs used by default in the scikit-learn display.
    user_style_kwargs : dict
        The user-defined Matplotlib style kwargs.

    Returns
    -------
    valid_style_kwargs : dict
        The validated style kwargs taking into account both default and user-defined
        Matplotlib style kwargs.
    """
    invalid_to_valid_kw = {
        "ls": "linestyle",
        "c": "color",
        "ec": "edgecolor",
        "fc": "facecolor",
        "lw": "linewidth",
        "mec": "markeredgecolor",
        "mfcalt": "markerfacecoloralt",
        "ms": "markersize",
        "mew": "markeredgewidth",
        "mfc": "markerfacecolor",
        "aa": "antialiased",
        "ds": "drawstyle",
        "font": "fontproperties",
        "family": "fontfamily",
        "name": "fontname",
        "size": "fontsize",
        "stretch": "fontstretch",
        "style": "fontstyle",
        "variant": "fontvariant",
        "weight": "fontweight",
        "ha": "horizontalalignment",
        "va": "verticalalignment",
        "ma": "multialignment",
    }
    for invalid_key, valid_key in invalid_to_valid_kw.items():
        if invalid_key in user_style_kwargs and valid_key in user_style_kwargs:
            raise TypeError(
                f"Got both {invalid_key} and {valid_key}, which are aliases of one "
                "another"
            )
    valid_style_kwargs = default_style_kwargs.copy()

    for key in user_style_kwargs:
        if key in invalid_to_valid_kw:
            valid_style_kwargs[invalid_to_valid_kw[key]] = user_style_kwargs[key]
        else:
            valid_style_kwargs[key] = user_style_kwargs[key]

    return valid_style_kwargs


def sample_mpl_colormap(
    cmap: Colormap, n: int
) -> list[tuple[float, float, float, float]]:
    """Sample colors from a Matplotlib colormap.

    Parameters
    ----------
    cmap : matplotlib.colors.Colormap
        The Matplotlib colormap to sample from.
    n : int
        The number of colors to sample.

    Returns
    -------
    colors : list of str
        The sampled colors.
    """
    indices = np.linspace(0, 1, n)
    return [cmap(i) for i in indices]



from collections.abc import Callable
from functools import wraps
from typing import Any

import matplotlib.pyplot as plt

DEFAULT_STYLE = {
    "font.size": 14,
    "axes.labelsize": 14,
    "axes.titlesize": 14,
    "xtick.labelsize": 13,
    "ytick.labelsize": 13,
    "legend.fontsize": 10,
    "legend.title_fontsize": 11,
    "axes.linewidth": 1.25,
    "grid.linewidth": 1.25,
    "lines.linewidth": 1.75,
    "lines.markersize": 6,
    "patch.linewidth": 1.25,
    "xtick.major.width": 1.5,
    "ytick.major.width": 1.5,
    "xtick.minor.width": 1.25,
    "ytick.minor.width": 1.25,
    "xtick.major.size": 7,
    "ytick.major.size": 7,
    "xtick.minor.size": 5,
    "ytick.minor.size": 5,
    "legend.loc": "upper left",
    "legend.borderaxespad": 0,
}


class StyleDisplayMixin:
    """Mixin to control the style plot of a display."""

    @property
    def _style_params(self) -> list[str]:
        """Get the list of available style parameters.

        Returns
        -------
        list
            List of style parameter names (without '_default_' prefix).
        """
        prefix = "_default_"
        suffix = "_kwargs"
        return [
            attr[len(prefix) :]
            for attr in dir(self)
            if attr.startswith(prefix) and attr.endswith(suffix)
        ]

    def set_style(self, **kwargs: Any):
        """Set the style parameters for the display.

        Parameters
        ----------
        **kwargs : dict
            Style parameters to set. Each parameter name should correspond to a
            a style attribute passed to the plot method of the display.

        Returns
        -------
        self : object
            Returns the instance itself.

        Raises
        ------
        ValueError
            If a style parameter is unknown.
        """
        for param_name, param_value in kwargs.items():
            default_attr = f"_default_{param_name}"
            if not hasattr(self, default_attr):
                raise ValueError(
                    f"Unknown style parameter: {param_name}. "
                    f"The parameter name should be one of {self._style_params}."
                )
            setattr(self, default_attr, param_value)
        return self

    @staticmethod
    def style_plot(plot_func: Callable) -> Callable:
        """Apply consistent style to skore displays.

        This decorator:
        1. Applies default style settings
        2. Executes `plot_func`
        3. Applies `tight_layout`

        Parameters
        ----------
        plot_func : callable
            The plot function to be decorated.

        Returns
        -------
        callable
            The decorated plot function.
        """

        @wraps(plot_func)
        def wrapper(self, *args: Any, **kwargs: Any) -> Any:
            # We need to manually handle setting the style of the parameters because
            # `plt.style.context` has a side effect with the interactive mode.
            # See https://github.com/matplotlib/matplotlib/issues/25041
            original_params = {key: plt.rcParams[key] for key in DEFAULT_STYLE}
            plt.rcParams.update(DEFAULT_STYLE)
            try:
                result = plot_func(self, *args, **kwargs)
            finally:
                plt.rcParams.update(original_params)
            return result

        return wrapper



"""Widget for interactive parallel coordinate plots of ML experiment metadata."""

from typing import Any, cast

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from IPython.display import clear_output, display
from ipywidgets import widgets
from rich.panel import Panel


class ModelExplorerWidget:
    """
    Widget for interactive parallel coordinate plots of ML experiment metadata.

    This class creates and manages interactive widgets and parallel coordinate plots
    for visually exploring machine learning experiment results. It allows users to
    filter and compare experiments across different metrics, learners, and datasets.

    Parameters
    ----------
    dataframe : pd.DataFrame
        The dataframe containing the experiment metadata with columns for metrics,
        learners, datasets, and ML task types.
    seed : int, default=0
        Random seed for jittering categorical columns to improve visualization.

    Attributes
    ----------
    current_fig : go.FigureWidget or None
        The currently displayed plotly figure.
    current_selection : dict
        Dictionary containing the current user selection criteria.
    """

    _plot_width: int = 800
    _metrics: dict[str, dict[str, str | bool]] = {
        "fit_time": {
            "name": "Fit Time",
            "greater_is_better": False,
            "type": "time",
            "show": False,
        },
        "predict_time": {
            "name": "Predict Time",
            "greater_is_better": False,
            "type": "time",
            "show": False,
        },
        "rmse": {
            "name": "RMSE",
            "greater_is_better": False,
            "type": "regression",
            "show": True,
        },
        "log_loss": {
            "name": "Log Loss",
            "greater_is_better": False,
            "type": "classification",
            "show": True,
        },
        "roc_auc": {
            "name": "Macro ROC AUC",
            "greater_is_better": True,
            "type": "classification",
            "show": True,
        },
    }
    _estimators: dict[str, dict[str, Any]] = {
        "learner": {"name": "Learner"},
    }
    _dimension_to_column: dict[str, str] = {
        cast(str, v["name"]): k for k, v in {**_metrics, **_estimators}.items()
    }

    _required_columns: list[str] = (
        ["ml_task", "dataset"] + list(_estimators.keys()) + list(_metrics.keys())
    )
    _required_index: list[str | None] = [None, "id"]

    def _check_dataframe_schema(self, dataframe: pd.DataFrame) -> None:
        """Check if the dataframe has the required columns and index."""
        if not all(col in dataframe.columns for col in self._required_columns):
            raise ValueError(
                f"Dataframe is missing required columns: {self._required_columns}"
            )
        if not all(col in dataframe.index.names for col in self._required_index):
            raise ValueError(
                f"Dataframe is missing required index: {self._required_index}"
            )
        if dataframe["learner"].dtype != "category":
            raise ValueError("Learner column must be a categorical column")

    def __init__(self, dataframe: pd.DataFrame, seed: int = 0) -> None:
        if dataframe.empty:
            self.dataframe = dataframe
            self.seed = seed
            return None

        self._check_dataframe_schema(dataframe)
        self.dataframe = dataframe
        self.seed = seed

        self.current_fig: go.FigureWidget | None = None
        self.current_selection: dict[str, Any] = {}

        self._clf_datasets: np.ndarray = self.dataframe.query(
            "ml_task.str.contains('classification')"
        )["dataset"].unique()
        self._reg_datasets: np.ndarray = self.dataframe.query(
            "ml_task.str.contains('regression')"
        )["dataset"].unique()

        default_task = "classification" if len(self._clf_datasets) else "regression"
        self._task_dropdown = widgets.Dropdown(
            options=[
                ("Classification", "classification"),
                ("Regression", "regression"),
            ],
            value=default_task,
            description="Task Type:",
            disabled=False,
            layout=widgets.Layout(width="200px"),
        )

        default_dataset = (
            self._clf_datasets
            if default_task == "classification"
            else self._reg_datasets
        )
        self._dataset_dropdown = widgets.Dropdown(
            options=default_dataset,
            description="Dataset:",
            disabled=False,
            layout=widgets.Layout(width="250px"),
        )

        self._metric_checkboxes: dict[str, dict[str, widgets.Checkbox]] = {
            "classification": {},
            "regression": {},
        }
        for metric in self._metrics:
            default_value = self._metrics[metric]["show"]
            metric_type = cast(str, self._metrics[metric]["type"])
            if metric_type == "time":
                # the "time" metrics should be added to all the different types
                # (i.e. classification and regression)
                for metric_type in self._metric_checkboxes:
                    self._metric_checkboxes[metric_type][metric] = widgets.Checkbox(
                        indent=False,
                        value=default_value,
                        description=cast(str, self._metrics[metric]["name"]),
                        disabled=False,
                        layout=widgets.Layout(width="auto", margin="0px 10px 0px 0px"),
                    )
            else:
                self._metric_checkboxes[metric_type][metric] = widgets.Checkbox(
                    indent=False,
                    value=default_value,
                    description=cast(str, self._metrics[metric]["name"]),
                    disabled=False,
                    layout=widgets.Layout(width="auto", margin="0px 10px 0px 0px"),
                )

        metrics_for_classification = [
            metric
            for metric in self._metrics
            if self._metrics[metric]["type"] in ("classification", "time")
        ]
        metrics_for_regression = [
            metric
            for metric in self._metrics
            if self._metrics[metric]["type"] in ("regression", "time")
        ]
        self._color_metric_dropdown: dict[str, widgets.Dropdown] = {
            "classification": widgets.Dropdown(
                options=[
                    cast(str, self._metrics[metric]["name"])
                    for metric in metrics_for_classification
                ],
                value="Log Loss",
                description="Color by:",
                disabled=False,
                layout=widgets.Layout(width="200px"),
            ),
            "regression": widgets.Dropdown(
                options=[
                    cast(str, self._metrics[metric]["name"])
                    for metric in metrics_for_regression
                ],
                value="RMSE",
                description="Color by:",
                disabled=False,
                layout=widgets.Layout(width="200px"),
            ),
        }
        self.classification_metrics_box = widgets.HBox(
            [
                self._metric_checkboxes["classification"][metric]
                for metric in metrics_for_classification
            ]
        )
        self.regression_metrics_box = widgets.HBox(
            [
                self._metric_checkboxes["regression"][metric]
                for metric in metrics_for_regression
            ]
        )

        controls_header = widgets.GridBox(
            [
                self._task_dropdown,
                self._dataset_dropdown,
                self._color_metric_dropdown["classification"],
                self._color_metric_dropdown["regression"],
            ],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_columns="repeat(4, auto)",
                grid_gap="5px",
                align_items="center",
            ),
        )

        clf_computation_row = widgets.GridBox(
            [
                widgets.Label(
                    value="Computation Metrics: ",
                    layout=widgets.Layout(padding="5px 0px"),
                ),
                self._metric_checkboxes["classification"]["fit_time"],
                self._metric_checkboxes["classification"]["predict_time"],
            ],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_columns="200px auto auto",
                align_items="center",
            ),
        )
        clf_statistical_row = widgets.GridBox(
            [
                widgets.Label(
                    value="Statistical Metrics: ",
                    layout=widgets.Layout(padding="5px 0px"),
                ),
                self._metric_checkboxes["classification"]["roc_auc"],
                self._metric_checkboxes["classification"]["log_loss"],
            ],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_columns="200px auto auto auto",
                align_items="center",
            ),
        )
        self.classification_metrics_box = widgets.GridBox(
            [clf_computation_row, clf_statistical_row],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_rows="auto auto",
                grid_gap="5px",
                align_items="center",
            ),
        )
        reg_computation_row = widgets.GridBox(
            [
                widgets.Label(
                    value="Computation Metrics: ",
                    layout=widgets.Layout(padding="5px 0px"),
                ),
                self._metric_checkboxes["regression"]["fit_time"],
                self._metric_checkboxes["regression"]["predict_time"],
            ],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_columns="200px auto auto",
                align_items="center",
            ),
        )
        reg_statistical_row = widgets.GridBox(
            [
                widgets.Label(
                    value="Statistical Metrics: ",
                    layout=widgets.Layout(padding="5px 0px"),
                ),
                self._metric_checkboxes["regression"]["rmse"],
            ],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_columns="200px auto auto",
                align_items="center",
            ),
        )
        self.regression_metrics_box = widgets.GridBox(
            [reg_computation_row, reg_statistical_row],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_rows="auto auto",
                grid_gap="5px",
                align_items="center",
            ),
        )
        controls_metrics = widgets.GridBox(
            [self.classification_metrics_box, self.regression_metrics_box],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_rows="auto auto",
                grid_gap="10px",
                align_items="center",
            ),
        )
        controls = widgets.GridBox(
            [controls_header, controls_metrics],
            layout=widgets.Layout(
                width=f"{self._plot_width}px",
                grid_template_rows="auto auto",
                grid_gap="10px",
                align_items="center",
                margin="0px 0px 5px 0px",
            ),
        )

        # callbacks
        self._task_dropdown.observe(self._on_task_change, names="value")
        self._dataset_dropdown.observe(self._update_plot, names="value")
        for task in self._metric_checkboxes:
            for metric in self._metric_checkboxes[task]:
                self._metric_checkboxes[task][metric].observe(
                    self._update_plot, names="value"
                )
            self._color_metric_dropdown[task].observe(self._update_plot, names="value")

        self.output = widgets.Output(
            layout=widgets.Layout(width=f"{self._plot_width}px", margin="0px")
        )

        self._update_task_widgets()
        self._layout = widgets.VBox(
            [controls, self.output],
            layout=widgets.Layout(width=f"{self._plot_width}px"),
        )

    def _update_task_widgets(self) -> None:
        """Update widget visibility based on the currently selected task."""
        task = self._task_dropdown.value

        if task == "classification":
            self.classification_metrics_box.layout.display = ""
            self._color_metric_dropdown["classification"].layout.display = ""

            self.regression_metrics_box.layout.display = "none"
            self._color_metric_dropdown["regression"].layout.display = "none"
        else:
            self.classification_metrics_box.layout.display = "none"
            self._color_metric_dropdown["classification"].layout.display = "none"

            self.regression_metrics_box.layout.display = ""
            self._color_metric_dropdown["regression"].layout.display = ""

    def _on_task_change(self, change: dict[str, Any]) -> None:
        """Handle task dropdown change events.

        Updates the dataset dropdown options based on the selected task
        and refreshes the widget visibility and plot.

        Parameters
        ----------
        change : dict[str, Any]
            dictionary containing information about the widget change,
            including the new value under the 'new' key.
        """
        task = change["new"]

        if task == "classification":
            self._dataset_dropdown.options = self._clf_datasets
            if len(self._clf_datasets):
                self._dataset_dropdown.value = self._clf_datasets[0]
        else:
            self._dataset_dropdown.options = self._reg_datasets
            if len(self._reg_datasets):
                self._dataset_dropdown.value = self._reg_datasets[0]

        self._update_task_widgets()
        self._update_plot()
        self.update_selection()

    @staticmethod
    def _add_jitter_to_categorical(
        seed: int, categorical_series: pd.Series, amount: float = 0.01
    ) -> np.ndarray:
        """Add jitter to categorical values to improve visualization in parallel plots.

        Jitter is not applied when there is a single category.

        Parameters
        ----------
        seed : int
            Random seed for reproducibility.
        categorical_series : pd.Series
            Categorical series to jitter.
        amount : float, default=0.01
            Amount of jitter to add.

        Returns
        -------
        np.ndarray
            Array of encoded categorical values with jitter applied.
        """
        if categorical_series.cat.categories.size == 1:
            return categorical_series.cat.codes.to_numpy()

        rng = np.random.default_rng(seed)
        encoded_categories = categorical_series.cat.codes.to_numpy()
        jitter = rng.uniform(-amount, amount, size=len(encoded_categories))
        for sign, cat in zip([1, -1], [0, len(encoded_categories) - 1], strict=False):
            jitter[encoded_categories == cat] = (
                np.abs(
                    jitter[encoded_categories == cat],
                    out=jitter[encoded_categories == cat],
                )
                * sign
            )

        return encoded_categories + jitter

    def _update_plot(self, change: dict[str, Any] | None = None) -> None:
        """Update the parallel coordinates plot based on the selected options.

        Creates a new plotly figure with dimensions for the selected metrics
        and displays it in the output area.

        Parameters
        ----------
        change : dict, default=None
            dictionary containing information about a widget change event.
            Not used directly but required for widget callback compatibility.
        """
        with self.output:
            clear_output(wait=True)

            task = self._task_dropdown.value

            if (
                not hasattr(self._dataset_dropdown, "value")
                or not self._dataset_dropdown.value
            ):
                display(widgets.HTML("No dataset available for selected task."))
                return

            df_dataset = self.dataframe.query(
                "dataset == @self._dataset_dropdown.value"
            ).copy()
            for col in df_dataset.select_dtypes(include=["category"]).columns:
                df_dataset[col] = df_dataset[col].cat.remove_unused_categories()

            selected_metrics = [
                metric
                for metric in self._metric_checkboxes[task]
                if self._metric_checkboxes[task][metric].value
            ]
            color_metric = self._dimension_to_column[
                self._color_metric_dropdown[task].value
            ]

            dimensions = []
            dimensions.append(
                dict(
                    label="Learner",
                    values=self._add_jitter_to_categorical(
                        self.seed, df_dataset["learner"]
                    ),
                    ticktext=df_dataset["learner"].cat.categories,
                    tickvals=np.arange(len(df_dataset["learner"].cat.categories)),
                )
            )

            for col in selected_metrics:  # use the order defined in the constructor
                dimensions.append(
                    dict(
                        label=cast(str, self._metrics[col]["name"]),
                        values=df_dataset[col].fillna(0),
                    )
                )

            colorscale = (
                "Viridis"
                if cast(bool, self._metrics[color_metric]["greater_is_better"])
                else "Viridis_r"
            )
            fig = go.FigureWidget(
                data=go.Parcoords(
                    line=dict(
                        color=df_dataset[color_metric].fillna(0),
                        colorscale=colorscale,
                        showscale=True,
                        colorbar=dict(
                            title=cast(str, self._metrics[color_metric]["name"])
                        ),
                    ),
                    dimensions=dimensions,
                    labelangle=-30,
                )
            )

            fig.update_layout(
                font=dict(size=16),
                height=500,
                width=self._plot_width,
                margin=dict(l=250, r=150, t=120, b=30),
                autosize=False,
            )

            fig.data[0].on_selection(self.update_selection)  # callback

            self.current_fig = fig
            display(fig)

    def update_selection(
        self, trace=None, points=None, selector=None
    ) -> "ModelExplorerWidget":
        """Update the current_selection attribute with the current filter state.

        Parameters
        ----------
        trace : trace, default=None
            The trace that triggered the selection change.
        points : list, default=None
            The points affected by the selection change.
        selector : dict, default=None
            The selector that triggered the selection change.

        Returns
        -------
        ModelExplorerWidget
            Self for method chaining.
        """
        selection_data = {
            "ml_task": self._task_dropdown.value,
            "dataset": self._dataset_dropdown.value,
        }

        if self.current_fig is not None:
            selection_data.update(
                {
                    self._dimension_to_column[dim.label]: dim.constraintrange
                    for dim in self.current_fig.data[0].dimensions
                    if hasattr(dim, "constraintrange") and dim.constraintrange
                }
            )

        self.current_selection = selection_data

        return self

    def display(self) -> None:
        """Display the widget interface and initialize the plot."""
        if self.dataframe.empty:
            from skore import console  # avoid circular import

            content = (
                "No report found in the project. Use the `put` method to add reports."
            )
            console.print(
                Panel(
                    content,
                    title="[bold cyan]Empty Project Metadata[/bold cyan]",
                    expand=False,
                    border_style="orange1",
                )
            )
            return None

        display(self._layout)
        self._update_plot()
        self.update_selection()



"""A helper to split a scikit-learn pipeline into estimators."""

import re
from collections import defaultdict

from sklearn.pipeline import FeatureUnion, Pipeline


def find_estimators(pipeline: Pipeline) -> dict[str, list[str]]:
    """
    Split the pipeline into a list of estimators or transformers.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline to split.

    Returns
    -------
    dict[str, list[str]]
        A dict of estimators or transformers class in the pipeline,
        grouped by their module name.

    Examples
    --------
    # with a simple scikit-learn pipeline
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.pipeline import make_pipeline
    >>> find_estimators(make_pipeline(LogisticRegression()))

    # with a skrub pipeline
    >>> from skrub import TableVectorizer
    >>> find_estimators(TableVectorizer())
    """
    itemized_pipeline = []
    from sklearn.base import BaseEstimator

    for _, value in pipeline.get_params().items():
        if isinstance(value, BaseEstimator) and not isinstance(value, FeatureUnion):
            itemized_pipeline.append(value)

    classes = set(type(est) for est in itemized_pipeline)

    sklearn_regex = re.compile(r"class \'sklearn\.(\w+)\.")
    skrub_regex = re.compile(r"class \'skrub\.(\w+)\.")

    pipe_steps: dict[str, list[str]] = defaultdict(list)
    for estimator_class in classes:
        class_name = estimator_class.__name__
        estimator_class_str = str(estimator_class)
        sklearn_match = sklearn_regex.search(estimator_class_str)
        skrub_match = skrub_regex.search(estimator_class_str)
        if sklearn_match is not None:
            module = sklearn_match.group(1)
        elif skrub_match is not None:
            module = skrub_match.group(1)
        else:
            module = "other"
        pipe_steps[module].append(class_name)

    # sort, so that even if two pipelines using the same transformers
    # are split in different orders, their dict still match.
    for _, value in pipe_steps.items():
        value.sort()

    return pipe_steps



from skore._sklearn._plot.data import TableReportDisplay
from skore._sklearn._plot.metrics import (
    ConfusionMatrixDisplay,
    MetricsSummaryDisplay,
    PrecisionRecallCurveDisplay,
    PredictionErrorDisplay,
    RocCurveDisplay,
)

__all__ = [
    "ConfusionMatrixDisplay",
    "RocCurveDisplay",
    "PrecisionRecallCurveDisplay",
    "PredictionErrorDisplay",
    "TableReportDisplay",
    "MetricsSummaryDisplay",
]



import pandas as pd
import skrub
from sklearn.utils.fixes import parse_version
from skrub import _dataframe as sbd
from skrub import _join_utils
from skrub._dataframe._common import dispatch

pandas_version = parse_version(parse_version(pd.__version__).base_version)
skrub_version = parse_version(parse_version(skrub.__version__).base_version)


if skrub_version < parse_version("0.6.0"):

    @dispatch
    def concat(*dataframes, axis=0):
        raise NotImplementedError()

    @concat.specialize("pandas", argument_type="DataFrame")
    def _concat_pandas(*dataframes, axis=0):
        kwargs = {"copy": False} if pandas_version < parse_version("3.0") else {}
        if axis == 0:
            return pd.concat(dataframes, axis=0, ignore_index=True, **kwargs)
        else:  # axis == 1
            init_index = dataframes[0].index
            dataframes = [df.reset_index(drop=True) for df in dataframes]
            dataframes = _join_utils.make_column_names_unique(*dataframes)
            result = pd.concat(dataframes, axis=1, **kwargs)
            result.index = init_index
            return result

    @concat.specialize("polars", argument_type="DataFrame")
    def _concat_polars(*dataframes, axis=0):
        import polars as pl

        if axis == 0:
            return pl.concat(dataframes, how="diagonal_relaxed")
        else:  # axis == 1
            dataframes = _join_utils.make_column_names_unique(*dataframes)
            return pl.concat(dataframes, how="horizontal")

    sbd.concat = concat


@dispatch
def to_frame(col):
    """Convert a single Column to a DataFrame."""
    raise NotImplementedError()


@to_frame.specialize("pandas", argument_type="Column")
def _to_frame_pandas(col):
    return col.to_frame()


@to_frame.specialize("polars", argument_type="Column")
def _to_frame_polars(col):
    return col.to_frame()


@dispatch
def is_in(col, values):
    raise NotImplementedError()


@is_in.specialize("pandas", argument_type="Column")
def _is_in_pandas(col, values):
    return col.isin(values)


@is_in.specialize("polars", argument_type="Column")
def _is_in_polars(col, values):
    return col.is_in(values)


sbd.to_frame = to_frame
sbd.is_in = is_in



from __future__ import annotations

import time
from collections import Counter
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal, cast

import joblib
import numpy as np
from numpy.typing import ArrayLike

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseReport
from skore._sklearn._cross_validation.report import CrossValidationReport
from skore._sklearn._estimator.report import EstimatorReport
from skore._sklearn.types import _DEFAULT, PositiveLabel
from skore._utils._progress_bar import progress_decorator

if TYPE_CHECKING:
    from skore._sklearn._estimator.metrics_accessor import _MetricsAccessor

    ReportType = Literal["EstimatorReport", "CrossValidationReport"]


class ComparisonReport(_BaseReport, DirNamesMixin):
    """Report for comparing reports.

    This object can be used to compare several :class:`skore.EstimatorReport` instances,
    or several :class:`~skore.CrossValidationReport` instances.

    Refer to the :ref:`comparison_report` section of the user guide for more details.

    .. caution::
       Reports passed to `ComparisonReport` are not copied. If you pass
       a report to `ComparisonReport`, and then modify the report outside later, it
       will affect the report stored inside the `ComparisonReport` as well, which
       can lead to inconsistent results. For this reason, modifying reports after
       creation is strongly discouraged.

    Parameters
    ----------
    reports : list of reports or dict
        Reports to compare. If a dict, keys will be used to label the estimators;
        if a list, the labels are computed from the estimator class names.
        Expects at least two reports to compare, with the same test target.

    n_jobs : int, default=None
        Number of jobs to run in parallel. Training the estimators and computing
        the scores are parallelized.
        When accessing some methods of the `ComparisonReport`, the `n_jobs`
        parameter is used to parallelize the computation.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    Attributes
    ----------
    reports_ : list of :class:`~skore.EstimatorReport` or list of \
        :class:`~skore.CrossValidationReport`
        The compared reports.

    report_names_ : list of str
        The names of the compared estimators. If the names are not customized (i.e. the
        class names are used), a de-duplication process is used to make sure that the
        names are distinct.

    See Also
    --------
    skore.EstimatorReport
        Report for a fitted estimator.

    skore.CrossValidationReport
        Report for the cross-validation of an estimator.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from skore import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import ComparisonReport, EstimatorReport
    >>> X, y = make_classification(random_state=42)
    >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
    >>> estimator_1 = LogisticRegression()
    >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
    >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
    >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
    >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
    >>> report.report_names_
    ['LogisticRegression_1', 'LogisticRegression_2']
    >>> report = ComparisonReport(
    ...     {"model1": estimator_report_1, "model2": estimator_report_2}
    ... )
    >>> report.report_names_
    ['model1', 'model2']

    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import ComparisonReport, CrossValidationReport
    >>> X, y = make_classification(random_state=42)
    >>> estimator_1 = LogisticRegression()
    >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
    >>> report_1 = CrossValidationReport(estimator_1, X, y)
    >>> report_2 = CrossValidationReport(estimator_2, X, y)
    >>> report = ComparisonReport([report_1, report_2])
    >>> report = ComparisonReport({"model1": report_1, "model2": report_2})
    """

    _ACCESSOR_CONFIG: dict[str, dict[str, str]] = {
        "metrics": {"name": "metrics"},
    }
    metrics: _MetricsAccessor

    _reports_type: ReportType

    @staticmethod
    def _validate_reports(
        reports: list[EstimatorReport]
        | dict[str, EstimatorReport]
        | list[CrossValidationReport]
        | dict[str, CrossValidationReport],
    ) -> tuple[
        list[EstimatorReport] | list[CrossValidationReport],
        list[str],
        ReportType,
        PositiveLabel,
    ]:
        """Validate that reports are in the right format for comparison.

        Parameters
        ----------
        reports : list of reports or dict
            The reports to be validated.

        Returns
        -------
        list of EstimatorReport or list of CrossValidationReport
            The validated reports.
        list of str
            The report names, either taken from dict keys or computed from the estimator
            class names.
        {"EstimatorReport", "CrossValidationReport"}
            The inferred type of the reports that will be compared.
        int, float, bool, str or None
            The positive label used in the different reports.
        """
        if not isinstance(reports, Iterable):
            raise TypeError(
                f"Expected reports to be a list or dict; got {type(reports)}"
            )

        if len(reports) < 2:
            raise ValueError(
                f"Expected at least 2 reports to compare; got {len(reports)}"
            )

        if isinstance(reports, list):
            report_names = None
            reports_list = reports
        else:  # dict
            report_names = list(reports.keys())
            for key in report_names:
                if not isinstance(key, str):
                    raise TypeError(
                        f"Expected all report names to be strings; got {type(key)}"
                    )
            reports_list = cast(
                list[EstimatorReport] | list[CrossValidationReport],
                list(reports.values()),
            )

        reports_type: ReportType
        if all(isinstance(report, EstimatorReport) for report in reports_list):
            reports_list = cast(list[EstimatorReport], reports_list)
            reports_type = "EstimatorReport"

            test_dataset_hashes = {
                joblib.hash(report.y_test)
                for report in reports_list
                if report.y_test is not None
            }
            if len(test_dataset_hashes) > 1:
                raise ValueError(
                    "Expected all estimators to share the same test targets."
                )

        elif all(isinstance(report, CrossValidationReport) for report in reports_list):
            reports_list = cast(list[CrossValidationReport], reports_list)
            reports_type = "CrossValidationReport"
        else:
            raise TypeError(
                f"Expected list or dict of {EstimatorReport.__name__} "
                f"or list of dict of {CrossValidationReport.__name__}"
            )

        if len(set(id(report) for report in reports_list)) < len(reports_list):
            raise ValueError("Expected reports to be distinct objects")

        ml_tasks = {report._ml_task for report in reports_list}
        if len(ml_tasks) > 1:
            raise ValueError(
                f"Expected all estimators to have the same ML usecase; got {ml_tasks}"
            )

        if ml_tasks == {"binary-classification"}:
            pos_labels = {report.pos_label for report in reports_list}
            if len(pos_labels) > 1:
                raise ValueError(
                    "Expected all estimators to have the same positive label. "
                    f"Got {pos_labels}."
                )
            pos_label = pos_labels.pop()
        else:
            pos_label = None

        if report_names is None:
            deduped_report_names = _deduplicate_report_names(
                [report.estimator_name_ for report in reports_list]
            )
        else:
            deduped_report_names = report_names

        return reports_list, deduped_report_names, reports_type, pos_label

    def __init__(
        self,
        reports: list[EstimatorReport]
        | dict[str, EstimatorReport]
        | list[CrossValidationReport]
        | dict[str, CrossValidationReport],
        *,
        n_jobs: int | None = None,
    ) -> None:
        """
        ComparisonReport instance initializer.

        Notes
        -----
        We check that the estimator reports can be compared:
        - all reports are estimator reports,
        - all estimators are in the same ML use case,
        - all estimators have non-empty X_test and y_test,
        - all estimators have the same X_test and y_test.
        """
        self.reports_, self.report_names_, self._reports_type, self._pos_label = (
            ComparisonReport._validate_reports(reports)
        )

        self._progress_info: dict[str, Any] | None = None

        self.n_jobs = n_jobs
        self._rng = np.random.default_rng(time.time_ns())
        self._hash = self._rng.integers(
            low=np.iinfo(np.int64).min, high=np.iinfo(np.int64).max
        )
        self._cache: dict[tuple[Any, ...], Any] = {}
        self._ml_task = self.reports_[0]._ml_task

    def clear_cache(self) -> None:
        """Clear the cache.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        >>> report.clear_cache()
        >>> report._cache
        {}
        """
        for report in self.reports_:
            report.clear_cache()
        self._cache = {}

    @progress_decorator(description="Estimator predictions")
    def cache_predictions(
        self,
        response_methods: Literal[
            "auto", "predict", "predict_proba", "decision_function"
        ] = "auto",
        n_jobs: int | None = None,
    ) -> None:
        """Cache the predictions for sub-estimators reports.

        Parameters
        ----------
        response_methods : {"auto", "predict", "predict_proba", "decision_function"},\
                default="auto
            The methods to use to compute the predictions.

        n_jobs : int, default=None
            The number of jobs to run in parallel. If `None`, we use the `n_jobs`
            parameter when initializing the report.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        >>> report._cache
        {...}
        """
        if n_jobs is None:
            n_jobs = self.n_jobs

        assert self._progress_info is not None, (
            "The rich Progress class was not initialized."
        )
        progress = self._progress_info["current_progress"]
        main_task = self._progress_info["current_task"]

        total_estimators = len(self.reports_)
        progress.update(main_task, total=total_estimators)

        for report in self.reports_:
            # Share the parent's progress bar with child report
            report._progress_info = {"current_progress": progress}
            report.cache_predictions(response_methods=response_methods, n_jobs=n_jobs)
            progress.update(main_task, advance=1, refresh=True)

    def get_predictions(
        self,
        *,
        data_source: Literal["train", "test", "X_y"],
        response_method: Literal[
            "predict", "predict_proba", "decision_function"
        ] = "predict",
        X: ArrayLike | None = None,
        pos_label: PositiveLabel | None = _DEFAULT,
    ) -> list[ArrayLike]:
        """Get predictions from the underlying reports.

        This method has the advantage to reload from the cache if the predictions
        were already computed in a previous call.

        Parameters
        ----------
        data_source : {"test", "train", "X_y"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.
            - "X_y" : use the provided `X` and `y` to compute the metric.

        response_method : {"predict", "predict_proba", "decision_function"}, \
                default="predict"
            The response method to use to get the predictions.

        X : array-like of shape (n_samples, n_features), optional
            When `data_source` is "X_y", the input features on which to compute the
            response method.

        pos_label : int, float, bool, str or None, default=_DEFAULT
            The label to consider as the positive class when computing predictions in
            binary classification cases. By default, the positive class is set to the
            one provided when creating the report. If `None`, `estimator_.classes_[1]`
            is used as positive label.

            When `pos_label` is equal to `estimator_.classes_[0]`, it will be equivalent
            to `estimator_.predict_proba(X)[:, 0]` for `response_method="predict_proba"`
            and `-estimator_.decision_function(X)` for
            `response_method="decision_function"`.

        Returns
        -------
        list of np.ndarray of shape (n_samples,) or (n_samples, n_classes) or list of \
                such lists
            The predictions for each :class:`~skore.EstimatorReport` or
            :class:`~skore.CrossValidationReport`.

        Raises
        ------
        ValueError
            If the data source is invalid.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from skore import train_test_split
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import ComparisonReport, EstimatorReport
        >>> X, y = make_classification(random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=42, as_dict=True)
        >>> estimator_1 = LogisticRegression()
        >>> estimator_report_1 = EstimatorReport(estimator_1, **split_data)
        >>> estimator_2 = LogisticRegression(C=2)  # Different regularization
        >>> estimator_report_2 = EstimatorReport(estimator_2, **split_data)
        >>> report = ComparisonReport([estimator_report_1, estimator_report_2])
        >>> report.cache_predictions()
        >>> predictions = report.get_predictions(data_source="test")
        >>> print([split_predictions.shape for split_predictions in predictions])
        [(25,), (25,)]
        """
        return [
            report.get_predictions(
                data_source=data_source,
                response_method=response_method,
                X=X,
                pos_label=pos_label,
            )
            for report in self.reports_
        ]

    @property
    def pos_label(self) -> PositiveLabel | None:
        return self._pos_label

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_panel_title(self) -> str:
        return "[bold cyan]Tools to compare estimators[/bold cyan]"

    def _get_help_legend(self) -> str:
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self) -> str:
        """Return a string representation."""
        return f"{self.__class__.__name__}(...)"


def _deduplicate_report_names(report_names: list[str]) -> list[str]:
    """De-duplicate report names that appear several times.

    Leave the other report names alone.

    Parameters
    ----------
    report_names : list of str
        The list of report names to be checked.

    Returns
    -------
    list of str
        The de-duplicated list of report names.

    Examples
    --------
    >>> _deduplicate_report_names(['a', 'b'])
    ['a', 'b']
    >>> _deduplicate_report_names(['a', 'a'])
    ['a_1', 'a_2']
    >>> _deduplicate_report_names(['a', 'b', 'a'])
    ['a_1', 'b', 'a_2']
    >>> _deduplicate_report_names(['a', 'b', 'a', 'b'])
    ['a_1', 'b_1', 'a_2', 'b_2']
    >>> _deduplicate_report_names([])
    []
    >>> _deduplicate_report_names(['a'])
    ['a']
    """
    counts = Counter(report_names)
    if len(report_names) == len(counts):
        return report_names

    names = report_names.copy()
    seen: Counter = Counter()
    for i in range(len(names)):
        name = names[i]
        seen[name] += 1
        if counts[name] > 1:
            names[i] = f"{name}_{seen[name]}"
    return names



import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix

from skore._sklearn._plot.base import Display
from skore._sklearn._plot.style import StyleDisplayMixin
from skore._sklearn._plot.utils import PlotBackendMixin


class ConfusionMatrixDisplay(PlotBackendMixin, Display):
    """Display for confusion matrix.

    Parameters
    ----------
    confusion_matrix : ndarray of shape (n_classes, n_classes)
        Confusion matrix.

    display_labels : list of str, default=None
        Display labels for plot. If None, display labels are set from 0 to
        ``n_classes - 1``.

    include_values : bool, default=True
        Includes values in confusion matrix.

    normalize : {'true', 'pred', 'all'}, default=None
        Normalizes confusion matrix over the true (rows), predicted (columns)
        conditions or all the population. If None, confusion matrix will not be
        normalized.

    values_format : str, default=None
        Format specification for values in confusion matrix. If None, the format
        specification is 'd' or '.2g' whichever is shorter.

    Attributes
    ----------
    figure_ : matplotlib Figure
        Figure containing the confusion matrix.

    ax_ : matplotlib Axes
        Axes with confusion matrix.

    text_ : ndarray of shape (n_classes, n_classes), dtype=matplotlib Text or \
            None
        Array of matplotlib text elements containing the values in the
        confusion matrix.
    """

    @StyleDisplayMixin.style_plot
    def __init__(
        self,
        confusion_matrix,
        *,
        display_labels=None,
        include_values=True,
        normalize=None,
        values_format=None,
    ):
        self.confusion_matrix = confusion_matrix
        self.display_labels = display_labels
        self.include_values = include_values
        self.normalize = normalize
        self.values_format = values_format
        self.figure_ = None
        self.ax_ = None
        self.text_ = None

    def _plot_matplotlib(self, ax=None, *, cmap="Blues", colorbar=True, **kwargs):
        """Plot the confusion matrix.

        Parameters
        ----------
        ax : matplotlib axes, default=None
            Axes object to plot on. If None, a new figure and axes is created.

        cmap : str or matplotlib Colormap, default='Blues'
            Colormap used for confusion matrix.

        colorbar : bool, default=True
            Whether or not to add a colorbar to the plot.

        **kwargs : dict
            Additional keyword arguments to be passed to matplotlib's
            `ax.imshow`.

        Returns
        -------
        self : ConfusionMatrixDisplay
            Configured with the confusion matrix.
        """
        if self.normalize not in (None, "true", "pred", "all"):
            raise ValueError(
                "normalize must be one of None, 'true', 'pred', 'all'; "
                f"got {self.normalize!r}"
            )

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        cm = self.confusion_matrix
        n_classes = cm.shape[0]

        with np.errstate(all="ignore"):
            if self.normalize == "true":
                cm = cm / cm.sum(axis=1, keepdims=True)
            elif self.normalize == "pred":
                cm = cm / cm.sum(axis=0, keepdims=True)
            elif self.normalize == "all":
                cm = cm / cm.sum()
            else:  # None
                pass

        self.confusion_matrix = cm

        im = ax.imshow(cm, interpolation="nearest", cmap=cmap, **kwargs)
        if colorbar:
            fig.colorbar(im, ax=ax)

        if self.display_labels is None:
            self.display_labels = np.arange(n_classes)
        elif len(self.display_labels) != n_classes:
            raise ValueError(
                f"display_labels must have length equal to number of classes "
                f"({n_classes}), got {len(self.display_labels)}"
            )
        ax.set(
            xticks=np.arange(n_classes),
            yticks=np.arange(n_classes),
            xticklabels=self.display_labels,
            yticklabels=self.display_labels,
            ylabel="True label",
            xlabel="Predicted label",
        )
        plt.setp(ax.get_xticklabels(), rotation=0, ha="center")

        self.text_ = np.empty_like(cm, dtype=object)
        if self.include_values:
            fmt = self.values_format or (".2f" if self.normalize else "d")
            thresh = cm.max() / 2.0
            for i in range(n_classes):
                for j in range(n_classes):
                    txt = format(cm[i, j], fmt)
                    color = "white" if cm[i, j] > thresh else "black"
                    self.text_[i, j] = ax.text(
                        j, i, txt, ha="center", va="center", color=color
                    )

        ax.set_title("Confusion Matrix")
        fig.tight_layout()
        self.figure_, self.ax_ = fig, ax
        return self

    @classmethod
    def from_predictions(
        cls,
        y_true,
        y_pred,
        *,
        sample_weight=None,
        display_labels=None,
        include_values=True,
        normalize=None,
        values_format=None,
    ):
        """Create a confusion matrix display from predictions.

        Parameters
        ----------
        y_true : array-like of shape (n_samples,)
            True labels.

        y_pred : array-like of shape (n_samples,)
            Predicted labels, as returned by a classifier.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.

        display_labels : list of str, default=None
            Target names used for plotting. By default, labels will be inferred
            from y_true.

        include_values : bool, default=True
            Includes values in confusion matrix.

        normalize : {'true', 'pred', 'all'}, default=None
            Normalizes confusion matrix over the true (rows), predicted (columns)
            conditions or all the population. If None, confusion matrix will not be
            normalized.

        values_format : str, default=None
            Format specification for values in confusion matrix. If None, the format
            specification is 'd' or '.2g' whichever is shorter.

        Returns
        -------
        display : :class:`~sklearn.metrics.ConfusionMatrixDisplay`
            The confusion matrix display.
        """
        cm = sklearn_confusion_matrix(
            y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
        )

        if display_labels is None:
            display_labels = np.unique(np.concatenate([y_true, y_pred]))

        disp = cls(
            confusion_matrix=cm,
            display_labels=display_labels,
            include_values=include_values,
            normalize=normalize,
            values_format=values_format,
        )

        return disp

    def frame(self):
        """Return the confusion matrix as a dataframe.

        Returns
        -------
        frame : pandas.DataFrame
            The confusion matrix as a dataframe.
        """
        import pandas as pd

        cm = self.confusion_matrix
        if self.display_labels is None:
            display_labels = [f"Class {i}" for i in range(cm.shape[0])]
        else:
            display_labels = self.display_labels

        return pd.DataFrame(cm, index=display_labels, columns=display_labels)



from collections.abc import Callable
from typing import Any

from sklearn.pipeline import Pipeline


def _check_all_checks(checks: list[Callable]) -> Callable:
    def check(accessor: Any) -> bool:
        return all(check(accessor) for check in checks)

    return check


def _check_roc_auc(ml_task_and_methods: list[tuple[str, list[str]]]):
    def check(accessor: Any) -> bool:
        are_supported_cases = []
        for ml_task, methods in ml_task_and_methods:
            is_supported_ml_task = ml_task in accessor._parent._ml_task
            has_methods = any(
                hasattr(accessor._parent._estimator, method) for method in methods
            )
            are_supported_cases.append(is_supported_ml_task and has_methods)

        if not any(are_supported_cases):
            err_msg = (
                f"For the task {accessor._parent._ml_task}, the estimator does not "
                "provide the right prediction methods. The called function requires "
                "the following combinations:\n\n"
            )
            err_msg += "\n".join(
                f"- {ml_task} with {methods}"
                for ml_task, methods in ml_task_and_methods
            )

            raise AttributeError(err_msg)

        return True

    return check


########################################################################################
# Accessor related to `EstimatorReport`
########################################################################################


def _check_supported_ml_task(supported_ml_tasks: list[str]) -> Callable:
    def check(accessor: Any) -> bool:
        supported_task = any(
            task in accessor._parent._ml_task for task in supported_ml_tasks
        )

        if not supported_task:
            raise AttributeError(
                f"The {accessor._parent._ml_task} task is not a supported task by "
                f"function called. The supported tasks are {supported_ml_tasks}."
            )

        return True

    return check


def _check_estimator_has_method(method_name: str) -> Callable:
    def check(accessor: Any) -> bool:
        parent_estimator = accessor._parent.estimator_

        if hasattr(parent_estimator, method_name):
            return True

        raise AttributeError(
            f"Estimator {parent_estimator} is not a supported estimator by "
            f"the function called. The estimator should have a `{method_name}` "
            "method."
        )

    return check


def _check_has_coef() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the estimator has a `coef_` attribute."""
        parent_estimator = accessor._parent.estimator_
        estimator = (
            parent_estimator.steps[-1][1]
            if isinstance(parent_estimator, Pipeline)
            else parent_estimator
        )
        if hasattr(estimator, "coef_"):
            return True
        try:  # e.g. TransformedTargetRegressor()
            if hasattr(estimator.regressor_, "coef_"):
                return True
        except AttributeError as msg:
            if "object has no attribute 'regressor_'" not in str(msg):
                raise
        raise AttributeError(
            f"Estimator {parent_estimator} is not a supported estimator by "
            "the function called."
        )

    return check


def _check_has_feature_importances() -> Callable:
    def check(accessor: Any) -> bool:
        """Check if the estimator has a `feature_importances_` attribute."""
        parent_estimator = accessor._parent.estimator_
        estimator = (
            parent_estimator.steps[-1][1]
            if isinstance(parent_estimator, Pipeline)
            else parent_estimator
        )
        if hasattr(estimator, "feature_importances_"):
            return True
        raise AttributeError(
            f"Estimator {parent_estimator} is not a supported estimator by "
            "the function called."
        )

    return check


########################################################################################
# Accessor related to `CrossValidationReport`
########################################################################################


def _check_estimator_report_has_method(
    accessor_name: str,
    method_name: str,
) -> Callable:
    def check(accessor: Any) -> bool:
        estimator_report = accessor._parent.estimator_reports_[0]

        if not hasattr(estimator_report, accessor_name):
            raise AttributeError(
                f"Estimator report {estimator_report} does not have the "
                f"'{accessor_name}' accessor."
            )
        accessor = getattr(estimator_report, accessor_name)

        if hasattr(accessor, method_name):
            return True
        raise AttributeError(
            f"Estimator report {estimator_report} is not a supported estimator report "
            "by the function called. The estimator report should have a "
            f"`{method_name}` method."
        )

    return check
